import abc
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
)

from typing_extensions import Literal

from galaxy.managers.context import ProvidesHistoryContext
from galaxy.model import (
    GalaxySession,
    History,
    Role,
)
from galaxy.model.base import transaction


class WorkRequestContext(ProvidesHistoryContext):
    """Stripped down implementation of Galaxy web transaction god object for
    work request handling outside of web threads - uses mix-ins shared with
    GalaxyWebTransaction to provide app, user, and history context convenience
    methods - but nothing related to HTTP handling, mako views, etc....

    Things that only need app shouldn't be consuming trans - but there is a
    need for actions potentially tied to users and histories and  hopefully
    this can define that stripped down interface providing access to user and
    history information - but not dealing with web request and response
    objects.
    """

    def __init__(
        self,
        app,
        user=None,
        history=None,
        workflow_building_mode=False,
        url_builder=None,
        galaxy_session: Optional[GalaxySession] = None,
    ):
        self._app = app
        self.__user = user
        self.__user_current_roles: Optional[List[Role]] = None
        self.__history = history
        self._url_builder = url_builder
        self._short_term_cache: Dict[Tuple[str, ...], Any] = {}
        self.workflow_building_mode = workflow_building_mode
        self.galaxy_session = galaxy_session

    def set_cache_value(self, args: Tuple[str, ...], value: Any):
        self._short_term_cache[args] = value

    def get_cache_value(self, args: Tuple[str, ...], default: Any = None) -> Any:
        return self._short_term_cache.get(args, default)

    @property
    def app(self):
        return self._app

    @property
    def url_builder(self):
        return self._url_builder

    def get_history(self, create=False):
        return self.__history

    @property
    def history(self):
        return self.get_history()

    def get_user(self):
        """Return the current user if logged in or None."""
        return self.__user

    def get_current_user_roles(self):
        if self.__user_current_roles is None:
            self.__user_current_roles = super().get_current_user_roles()
        return self.__user_current_roles

    def set_user(self, user):
        """Set the current user."""
        raise NotImplementedError("Cannot change users from a work request context.")

    user = property(get_user, set_user)


class GalaxyAbstractRequest:
    """Abstract interface to provide access to some request properties."""

    @abc.abstractproperty
    def base(self) -> str:
        """Base URL of the request."""

    @abc.abstractproperty
    def host(self) -> str:
        """The host address."""

    @abc.abstractproperty
    def is_secure(self) -> bool:
        """Was this a secure (https) request."""

    @abc.abstractmethod
    def get_cookie(self, name):
        """Return cookie."""


class GalaxyAbstractResponse:
    """Abstract interface to provide access to some response utilities."""

    @abc.abstractproperty
    def headers(self) -> dict:
        """The response headers."""

    def set_content_type(self, content_type: str):
        """
        Sets the Content-Type header
        """
        self.headers["content-type"] = content_type

    def get_content_type(self):
        return self.headers.get("content-type", None)

    @abc.abstractmethod
    def set_cookie(
        self,
        key: str,
        value: str = "",
        max_age: Optional[int] = None,
        expires: Optional[int] = None,
        path: str = "/",
        domain: Optional[str] = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Optional[Literal["lax", "strict", "none"]] = "lax",
    ) -> None:
        """Set a cookie."""


class SessionRequestContext(WorkRequestContext):
    """Like WorkRequestContext, but provides access to request."""

    def __init__(self, **kwargs):
        self.__request: GalaxyAbstractRequest = kwargs.pop("request")
        self.__response: GalaxyAbstractResponse = kwargs.pop("response")
        super().__init__(**kwargs)

    @property
    def host(self):
        return self.__request.host

    @property
    def request(self) -> GalaxyAbstractRequest:
        return self.__request

    @property
    def response(self) -> GalaxyAbstractResponse:
        return self.__response

    def get_galaxy_session(self):
        return self.galaxy_session

    def set_history(self, history):
        if history and not history.deleted and self.galaxy_session:
            self.galaxy_session.current_history = history
        self.sa_session.add(self.galaxy_session)
        with transaction(self.sa_session):
            self.sa_session.commit()


def proxy_work_context_for_history(
    trans: ProvidesHistoryContext, history: Optional[History] = None, workflow_building_mode=False
):
    """Create a WorkContext for supplied context with potentially different history.

    This provides semi-structured access to a transaction/work context with a supplied target
    history that is different from the user's current history (which also might change during
    the request).
    """
    return WorkRequestContext(
        app=trans.app,
        user=trans.user,
        history=history or trans.history,
        url_builder=trans.url_builder,
        workflow_building_mode=workflow_building_mode,
        galaxy_session=trans.galaxy_session,
    )



"""
Once state information has been calculated, handle actually executing tools
from various states, tracking results, and building implicit dataset
collections from matched collections.
"""

import collections
import logging
import typing
from abc import abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
)

from boltons.iterutils import remap
from packaging.version import Version

from galaxy import model
from galaxy.exceptions import ToolInputsNotOKException
from galaxy.model.base import transaction
from galaxy.model.dataset_collections.matching import MatchingCollections
from galaxy.model.dataset_collections.structure import (
    get_structure,
    tool_output_to_structure,
)
from galaxy.tool_util.parser import ToolOutputCollectionPart
from galaxy.tools.actions import (
    filter_output,
    on_text_for_names,
    ToolExecutionCache,
)
from galaxy.tools.parameters.basic import is_runtime_value

if typing.TYPE_CHECKING:
    from galaxy.tools import Tool

log = logging.getLogger(__name__)

SINGLE_EXECUTION_SUCCESS_MESSAGE = "Tool ${tool_id} created job ${job_id}"
BATCH_EXECUTION_MESSAGE = "Created ${job_count} job(s) for tool ${tool_id} request"


class PartialJobExecution(Exception):
    def __init__(self, execution_tracker):
        self.execution_tracker = execution_tracker


class MappingParameters(NamedTuple):
    param_template: Dict[str, Any]
    param_combinations: List[Dict[str, Any]]


def execute(
    trans,
    tool: "Tool",
    mapping_params: MappingParameters,
    history: model.History,
    rerun_remap_job_id: Optional[int] = None,
    preferred_object_store_id: Optional[str] = None,
    collection_info: Optional[MatchingCollections] = None,
    workflow_invocation_uuid: Optional[str] = None,
    invocation_step: Optional[model.WorkflowInvocationStep] = None,
    max_num_jobs: Optional[int] = None,
    job_callback: Optional[Callable] = None,
    completed_jobs: Optional[Dict[int, Optional[model.Job]]] = None,
    workflow_resource_parameters: Optional[Dict[str, Any]] = None,
    validate_outputs: bool = False,
):
    """
    Execute a tool and return object containing summary (output data, number of
    failures, etc...).
    """
    completed_jobs = completed_jobs or {}
    if max_num_jobs is not None:
        assert invocation_step is not None
    if rerun_remap_job_id:
        assert invocation_step is None

    all_jobs_timer = tool.app.execution_timer_factory.get_timer(
        "internals.galaxy.tools.execute.job_batch", BATCH_EXECUTION_MESSAGE
    )

    if invocation_step is None:
        execution_tracker: ExecutionTracker = ToolExecutionTracker(
            trans, tool, mapping_params, collection_info, completed_jobs=completed_jobs
        )
    else:
        execution_tracker = WorkflowStepExecutionTracker(
            trans, tool, mapping_params, collection_info, invocation_step, completed_jobs=completed_jobs
        )
    execution_cache = ToolExecutionCache(trans)

    def execute_single_job(execution_slice, completed_job, skip=False):
        job_timer = tool.app.execution_timer_factory.get_timer(
            "internals.galaxy.tools.execute.job_single", SINGLE_EXECUTION_SUCCESS_MESSAGE
        )
        params = execution_slice.param_combination
        if "__data_manager_mode" in mapping_params.param_template:
            params["__data_manager_mode"] = mapping_params.param_template["__data_manager_mode"]
        if workflow_invocation_uuid:
            params["__workflow_invocation_uuid__"] = workflow_invocation_uuid
        elif "__workflow_invocation_uuid__" in params:
            # Only workflow invocation code gets to set this, ignore user supplied
            # values or rerun parameters.
            del params["__workflow_invocation_uuid__"]
        if workflow_resource_parameters:
            params["__workflow_resource_params__"] = workflow_resource_parameters
        elif "__workflow_resource_params__" in params:
            # Only workflow invocation code gets to set this, ignore user supplied
            # values or rerun parameters.
            del params["__workflow_resource_params__"]
        if validate_outputs:
            params["__validate_outputs__"] = True
        job, result = tool.handle_single_execution(
            trans,
            rerun_remap_job_id,
            execution_slice,
            history,
            execution_cache,
            completed_job,
            collection_info,
            job_callback=job_callback,
            preferred_object_store_id=preferred_object_store_id,
            flush_job=False,
            skip=skip,
        )
        if job:
            log.debug(job_timer.to_str(tool_id=tool.id, job_id=job.id))
            execution_tracker.record_success(execution_slice, job, result)
            # associate dataset instances with the job that creates them
            if result:
                instance_types = (model.HistoryDatasetAssociation, model.LibraryDatasetDatasetAssociation)
                datasets = [pair[1] for pair in result if type(pair[1]) in instance_types]
                if datasets:
                    job_datasets[job] = datasets
        else:
            execution_tracker.record_error(result)

    tool_action = tool.tool_action
    if check_inputs_ready := getattr(tool_action, "check_inputs_ready", None):
        for params in execution_tracker.param_combinations:
            # This will throw an exception if the tool is not ready.
            try:
                check_inputs_ready(
                    tool,
                    trans,
                    params,
                    history,
                    execution_cache=execution_cache,
                    collection_info=collection_info,
                )
            except ToolInputsNotOKException as e:
                execution_tracker.record_error(e)

    execution_tracker.ensure_implicit_collections_populated(history, mapping_params.param_template)
    job_count = len(execution_tracker.param_combinations)

    jobs_executed = 0
    has_remaining_jobs = False
    execution_slice = None
    job_datasets: Dict[str, List[model.DatasetInstance]] = {}  # job: list of dataset instances created by job

    for i, execution_slice in enumerate(execution_tracker.new_execution_slices()):
        if max_num_jobs is not None and jobs_executed >= max_num_jobs:
            has_remaining_jobs = True
            break
        else:
            skip = execution_slice.param_combination.pop("__when_value__", None) is False
            execute_single_job(execution_slice, completed_jobs[i], skip=skip)
            history = execution_slice.history or history
            jobs_executed += 1

    if execution_slice:
        history.add_pending_items()
    # Make sure collections, implicit jobs etc are flushed even if there are no precreated output datasets
    with transaction(trans.sa_session):
        trans.sa_session.commit()

    tool_id = tool.id
    for job2 in execution_tracker.successful_jobs:
        # Put the job in the queue if tracking in memory
        if tool_id == "__DATA_FETCH__" and tool.app.config.is_fetch_with_celery_enabled():
            job_id = job2.id
            from galaxy.celery.tasks import (
                fetch_data,
                finish_job,
                set_job_metadata,
                setup_fetch_data,
            )

            raw_tool_source = tool.tool_source.to_string()
            #  task_user_id parameter is used to do task user rate limiting. It is only passed
            #  to first task in chain because it is only necessary to rate limit the first
            #  task in a chain.
            async_result = (
                setup_fetch_data.s(
                    job_id, raw_tool_source=raw_tool_source, task_user_id=getattr(trans.user, "id", None)
                )
                | fetch_data.s(job_id=job_id)
                | set_job_metadata.s(
                    extended_metadata_collection="extended" in tool.app.config.metadata_strategy,
                    job_id=job_id,
                ).set(link_error=finish_job.si(job_id=job_id, raw_tool_source=raw_tool_source))
                | finish_job.si(job_id=job_id, raw_tool_source=raw_tool_source)
            )()
            job2.set_runner_external_id(async_result.task_id)
            continue
        tool.app.job_manager.enqueue(job2, tool=tool, flush=False)
        trans.log_event(f"Added job to the job queue, id: {str(job2.id)}", tool_id=tool_id)
    with transaction(trans.sa_session):
        trans.sa_session.commit()

    if has_remaining_jobs:
        raise PartialJobExecution(execution_tracker)
    else:
        execution_tracker.finalize_dataset_collections(trans)

    log.debug(all_jobs_timer.to_str(job_count=job_count, tool_id=tool.id))
    return execution_tracker


class ExecutionSlice:
    def __init__(self, job_index, param_combination, dataset_collection_elements=None):
        self.job_index = job_index
        self.param_combination = param_combination
        self.dataset_collection_elements = dataset_collection_elements
        self.history = None


class ExecutionTracker:
    def __init__(self, trans, tool, mapping_params, collection_info, completed_jobs=None):
        # Known ahead of time...
        self.trans = trans
        self.tool = tool
        self.mapping_params = mapping_params
        self.collection_info = collection_info
        self.completed_jobs = completed_jobs

        self._on_text = None

        # Populated as we go...
        self.failed_jobs = 0
        self.execution_errors = []

        self.successful_jobs = []
        self.output_datasets = []
        self.output_collections = []

        self.implicit_collections = {}

    @property
    def param_combinations(self):
        return self.mapping_params.param_combinations

    @property
    def example_params(self):
        if self.mapping_params.param_combinations:
            return self.mapping_params.param_combinations[0]
        else:
            # TODO: This isn't quite right - what we want is something like param_template wrapped,
            # need a test case with an output filter applied to an empty list, still this is
            # an improvement over not allowing mapping of empty lists.
            return self.mapping_params.param_template

    @property
    def job_count(self):
        return len(self.param_combinations)

    def record_error(self, error):
        self.failed_jobs += 1
        message = "There was a failure executing a job for tool [%s] - %s"
        log.warning(message, self.tool.id, error)
        self.execution_errors.append(error)

    @property
    def on_text(self):
        if self._on_text is None:
            collection_names = ["collection %d" % c.hid for c in self.collection_info.collections.values()]
            self._on_text = on_text_for_names(collection_names)

        return self._on_text

    def output_name(self, trans, history, params, output):
        on_text = self.on_text

        try:
            output_collection_name = self.tool.tool_action.get_output_name(
                output,
                dataset=None,
                tool=self.tool,
                on_text=on_text,
                trans=trans,
                history=history,
                params=params,
                incoming=None,
                job_params=None,
            )
        except Exception:
            output_collection_name = f"{self.tool.name} across {on_text}"

        return output_collection_name

    def sliced_input_collection_structure(self, input_name):
        unqualified_recurse = Version(str(self.tool.profile)) < Version("18.09") and "|" not in input_name

        def find_collection(input_dict, input_name):
            for key, value in input_dict.items():
                if key == input_name:
                    return value
                if isinstance(value, dict):
                    if "|" in input_name:
                        prefix, rest_input_name = input_name.split("|", 1)
                        if key == prefix:
                            return find_collection(value, rest_input_name)
                    elif unqualified_recurse:
                        # Looking for "input1" instead of "cond|input1" for instance.
                        # See discussion on https://github.com/galaxyproject/galaxy/issues/6157.
                        unqualified_match = find_collection(value, input_name)
                        if unqualified_match:
                            return unqualified_match

        input_collection = find_collection(self.example_params, input_name)
        if input_collection is None:
            raise Exception("Failed to find referenced collection in inputs.")

        if not hasattr(input_collection, "collection"):
            raise Exception("Referenced input parameter is not a collection.")

        collection_type_description = (
            self.trans.app.dataset_collection_manager.collection_type_descriptions.for_collection_type(
                input_collection.collection.collection_type
            )
        )
        subcollection_mapping_type = None
        if self.is_implicit_input(input_name):
            subcollection_mapping_type = self.collection_info.subcollection_mapping_type(input_name)

        return get_structure(
            input_collection, collection_type_description, leaf_subcollection_type=subcollection_mapping_type
        )

    def _structure_for_output(self, trans, tool_output):
        structure = self.collection_info.structure
        if hasattr(tool_output, "default_identifier_source"):
            # Switch the structure for outputs if the output specified a default_identifier_source
            collection_type_descriptions = trans.app.dataset_collection_manager.collection_type_descriptions

            source_collection = self.collection_info.collections.get(tool_output.default_identifier_source)
            if source_collection:
                collection_type_description = collection_type_descriptions.for_collection_type(
                    source_collection.collection.collection_type
                )
                _structure = structure.for_dataset_collection(
                    source_collection.collection, collection_type_description=collection_type_description
                )
                if structure.can_match(_structure):
                    structure = _structure

        return structure

    def _mapped_output_structure(self, trans, tool_output):
        collections_manager = trans.app.dataset_collection_manager
        output_structure = tool_output_to_structure(
            self.sliced_input_collection_structure, tool_output, collections_manager
        )
        # self.collection_info.structure - the mapping structure with default_identifier_source
        # used to determine the identifiers to use.
        mapping_structure = self._structure_for_output(trans, tool_output)
        # Output structure may not be known, but input structure must be,
        # otherwise this step of the workflow shouldn't have been scheduled
        # or the tool should not have been executable on this input.
        mapped_output_structure = mapping_structure.multiply(output_structure)
        return mapped_output_structure

    def ensure_implicit_collections_populated(self, history, params):
        if not self.collection_info:
            return

        history = history or self.tool.get_default_history_by_trans(self.trans)
        # params = param_combinations[0] if param_combinations else mapping_params.param_template
        self.precreate_output_collections(history, params)

    def precreate_output_collections(self, history, params):
        # params is just one sample tool param execution with parallelized
        # collection replaced with a specific dataset. Need to replace this
        # with the collection and wrap everything up so can evaluate output
        # label.
        trans = self.trans
        params.update(
            self.collection_info.collections
        )  # Replace datasets with source collections for labelling outputs.

        collection_instances = {}
        implicit_inputs = self.implicit_inputs

        implicit_collection_jobs = model.ImplicitCollectionJobs()

        # trying to guess these filters at the collection levell is tricky because
        # the filter condition could vary from element to element. Just do best we
        # we can for now.
        example_params = self.example_params.copy()

        # walk through and optional replace runtime values with None, assume they
        # would have been replaced by now if they were going to be set.
        def replace_optional_runtime_values(path, key, value):
            if is_runtime_value(value):
                return key, None
            return key, value

        example_params = remap(example_params, visit=replace_optional_runtime_values)

        for output_name, output in self.tool.outputs.items():
            if filter_output(self.tool, output, example_params):
                continue
            output_collection_name = self.output_name(trans, history, params, output)
            effective_structure = self._mapped_output_structure(trans, output)
            collection_instance = trans.app.dataset_collection_manager.precreate_dataset_collection_instance(
                trans=trans,
                parent=history,
                name=output_collection_name,
                structure=effective_structure,
                implicit_inputs=implicit_inputs,
                implicit_output_name=output_name,
                completed_collection=self.completed_jobs,
            )
            collection_instance.implicit_collection_jobs = implicit_collection_jobs
            collection_instances[output_name] = collection_instance
            trans.sa_session.add(collection_instance)
        # Needed to flush the association created just above with
        # job.add_output_dataset_collection.
        with transaction(trans.sa_session):
            trans.sa_session.commit()
        self.implicit_collections = collection_instances

    @property
    def implicit_collection_jobs(self):
        # TODO: refactor to track this properly maybe?
        if self.implicit_collections:
            return next(iter(self.implicit_collections.values())).implicit_collection_jobs
        else:
            return None

    def finalize_dataset_collections(self, trans):
        # TODO: this probably needs to be reworked some, we should have the collection methods
        # return a list of changed objects to add to the session and flush and we should only
        # be finalizing collections to a depth of self.collection_info.structure. So for instance
        # if you are mapping a list over a tool that dynamically generates lists - we won't actually
        # know the structure of the inner list until after its job is complete.
        if self.failed_jobs > 0:
            for i, implicit_collection in enumerate(self.implicit_collections.values()):
                if i == 0:
                    implicit_collection_jobs = implicit_collection.implicit_collection_jobs
                    implicit_collection_jobs.populated_state = "failed"
                    trans.sa_session.add(implicit_collection_jobs)
                implicit_collection.collection.handle_population_failed(
                    "One or more jobs failed during dataset initialization."
                )
                trans.sa_session.add(implicit_collection.collection)
        else:
            completed_collections = {}
            if (
                self.completed_jobs
                and self.implicit_collection_jobs
                and len(self.completed_jobs) == len(self.successful_jobs)
            ):
                # If the same number of implicit collection jobs went into
                # creating the collection and those jobs are all cached
                # the HDCA has effectively been copied.
                # We mark this here so that the job cache query in subsequent
                # jobs considers this to be a valid cached input.
                completed_job_ids = {job.id for job in self.completed_jobs.values() if job}
                if all(job.copied_from_job_id in completed_job_ids for job in self.implicit_collection_jobs.job_list):
                    completed_collections = {
                        jtodca.name: jtodca.dataset_collection_instance
                        for jtodca in self.completed_jobs[0].output_dataset_collection_instances
                    }
            implicit_collection = None
            for i, implicit_collection in enumerate(self.implicit_collections.values()):
                if i == 0:
                    implicit_collection_jobs = implicit_collection.implicit_collection_jobs
                    implicit_collection_jobs.populated_state = "ok"
                    trans.sa_session.add(implicit_collection_jobs)
                implicit_collection.collection.finalize(
                    collection_type_description=self.collection_info.structure.collection_type_description
                )

                # Mark implicit HDCA as copied
                completed_implicit_collection = implicit_collection and completed_collections.get(
                    implicit_collection.implicit_output_name
                )
                if completed_implicit_collection:
                    implicit_collection.copied_from_history_dataset_collection_association_id = (
                        completed_implicit_collection.id
                    )

                trans.sa_session.add(implicit_collection.collection)

        with transaction(trans.sa_session):
            trans.sa_session.commit()

    @property
    def implicit_inputs(self):
        implicit_inputs = list(self.collection_info.collections.items())
        return implicit_inputs

    def is_implicit_input(self, input_name):
        return input_name in self.collection_info.collections

    def walk_implicit_collections(self):
        return self.collection_info.structure.walk_collections(self.implicit_collections)

    def new_execution_slices(self):
        if self.collection_info is None:
            for job_index, param_combination in enumerate(self.param_combinations):
                yield ExecutionSlice(job_index, param_combination)
        else:
            yield from self.new_collection_execution_slices()

    def record_success(self, execution_slice, job, outputs):
        # TODO: successful_jobs need to be inserted in the correct place...
        self.successful_jobs.append(job)
        self.output_datasets.extend(outputs)
        for job_output in job.output_dataset_collection_instances:
            self.output_collections.append((job_output.name, job_output.dataset_collection_instance))
        if self.implicit_collections:
            implicit_collection_jobs = None
            for output_name, collection_instance in self.implicit_collections.items():
                job.add_output_dataset_collection(output_name, collection_instance)
                if implicit_collection_jobs is None:
                    implicit_collection_jobs = collection_instance.implicit_collection_jobs

            job_assoc = model.ImplicitCollectionJobsJobAssociation()
            job_assoc.order_index = execution_slice.job_index
            job_assoc.implicit_collection_jobs = implicit_collection_jobs
            job_assoc.job = job
            self.trans.sa_session.add(job_assoc)

    @abstractmethod
    def new_collection_execution_slices(self):
        pass


# Seperate these because workflows need to track their jobs belong to the invocation
# in the database immediately and they can be recovered.
class ToolExecutionTracker(ExecutionTracker):
    def __init__(self, trans, tool, mapping_params, collection_info, completed_jobs=None):
        super().__init__(trans, tool, mapping_params, collection_info, completed_jobs=completed_jobs)

        # New to track these things for tool output API response in the tool case,
        # in the workflow case we just write stuff to the database and forget about
        # it.
        self.outputs_by_output_name = collections.defaultdict(list)

    def record_success(self, execution_slice, job, outputs):
        super().record_success(execution_slice, job, outputs)
        for output_name, output_dataset in outputs:
            if ToolOutputCollectionPart.is_named_collection_part_name(output_name):
                # Skip known collection outputs, these will be covered by
                # output collections.
                continue
            self.outputs_by_output_name[output_name].append(output_dataset)
        for job_output in job.output_dataset_collections:
            self.outputs_by_output_name[job_output.name].append(job_output.dataset_collection)

    def new_collection_execution_slices(self):
        for job_index, (param_combination, (dataset_collection_elements, _when_value)) in enumerate(
            zip(self.param_combinations, self.walk_implicit_collections())
        ):
            completed_job = self.completed_jobs and self.completed_jobs[job_index]
            if not completed_job:
                for dataset_collection_element in dataset_collection_elements.values():
                    assert dataset_collection_element.element_object is None

            yield ExecutionSlice(job_index, param_combination, dataset_collection_elements)


class WorkflowStepExecutionTracker(ExecutionTracker):
    def __init__(self, trans, tool, mapping_params, collection_info, invocation_step, completed_jobs=None):
        super().__init__(trans, tool, mapping_params, collection_info, completed_jobs=completed_jobs)
        self.invocation_step = invocation_step

    def record_success(self, execution_slice, job, outputs):
        super().record_success(execution_slice, job, outputs)
        if not self.collection_info:
            for output_name, output in outputs:
                self.invocation_step.add_output(output_name, output)
            self.invocation_step.job = job

    def new_collection_execution_slices(self):
        for job_index, (param_combination, (dataset_collection_elements, _when_value)) in enumerate(
            zip(self.param_combinations, self.walk_implicit_collections())
        ):
            completed_job = self.completed_jobs and self.completed_jobs[job_index]
            if not completed_job:
                found_result = False
                for dataset_collection_element in dataset_collection_elements.values():
                    if dataset_collection_element.element_object is not None:
                        found_result = True
                        break
                if found_result:
                    continue

            yield ExecutionSlice(job_index, param_combination, dataset_collection_elements)

    def ensure_implicit_collections_populated(self, history, params):
        if not self.collection_info:
            return

        history = history or self.tool.get_default_history_by_trans(self.trans)
        if self.invocation_step.is_new:
            self.precreate_output_collections(history, params)
            for output_name, implicit_collection in self.implicit_collections.items():
                self.invocation_step.add_output(output_name, implicit_collection)
        else:
            collections = {}
            for output_assoc in self.invocation_step.output_dataset_collections:
                implicit_collection = output_assoc.dataset_collection
                assert hasattr(implicit_collection, "history_content_type")  # make sure it is an HDCA and not a DC
                collections[output_assoc.output_name] = output_assoc.dataset_collection
            self.implicit_collections = collections
        self.invocation_step.implicit_collection_jobs = self.implicit_collection_jobs


__all__ = ("execute",)



"""Utilities for loading tools and workflows from paths for admin user requests."""

from typing import (
    Any,
    Dict,
    Optional,
)

import yaml

from galaxy import exceptions
from galaxy.util import in_directory


def artifact_class(trans, as_dict: Dict[str, Any], allow_in_directory: Optional[str] = None):
    object_id = as_dict.get("object_id", None)
    if as_dict.get("src", None) == "from_path":
        workflow_path = as_dict.get("path")
        allow = not trans or trans.user_is_admin
        allow = allow or (allow_in_directory and in_directory(workflow_path, allow_in_directory))

        if not allow:
            raise exceptions.AdminRequiredException()

        if not isinstance(workflow_path, str):
            raise exceptions.RequestParameterInvalidException()

        with open(workflow_path) as f:
            as_dict = yaml.safe_load(f)

    artifact_class = as_dict.get("class", None)
    if artifact_class is None and "$graph" in as_dict:
        object_id = object_id or "main"
        graph = as_dict["$graph"]
        target_object = None
        if isinstance(graph, dict):
            target_object = graph.get(object_id)
        else:
            for item in graph:
                found_id = item.get("id")
                if found_id == object_id or found_id == f"#{object_id}":
                    target_object = item

        if target_object and target_object.get("class"):
            artifact_class = target_object["class"]

    return artifact_class, as_dict, object_id


__all__ = ("artifact_class",)



import errno
import logging
import os
from typing import (
    Dict,
    List,
    Optional,
    Union,
)

from typing_extensions import Protocol

from galaxy import util
from galaxy.structured_app import StructuredApp
from galaxy.tool_shed.galaxy_install.client import DataManagersInterface
from galaxy.tool_util.data import (
    BundleProcessingOptions,
    OutputDataset,
)
from galaxy.tool_util.data.bundles.models import (
    convert_data_tables_xml,
    RepoInfo,
)
from galaxy.util import Element

log = logging.getLogger(__name__)


class DataManagers(DataManagersInterface):
    data_managers: Dict[str, "DataManager"]
    managed_data_tables: Dict[str, "DataManager"]
    __reload_count: int

    def __init__(self, app: StructuredApp, xml_filename=None, reload_count: Optional[int] = None):
        self.app = app
        self.data_managers = {}
        self.managed_data_tables = {}
        self.tool_path: Optional[str] = None
        self.__reload_count = reload_count or 0
        self.filename = xml_filename or self.app.config.data_manager_config_file
        for filename in util.listify(self.filename):
            if not filename:
                continue
            self.load_from_xml(filename)
        if self.app.config.shed_data_manager_config_file:
            try:
                self.load_from_xml(self.app.config.shed_data_manager_config_file, store_tool_path=True)
            except OSError as exc:
                if exc.errno != errno.ENOENT or self.app.config.is_set("shed_data_manager_config_file"):
                    raise

    @property
    def _reload_count(self) -> int:
        return self.__reload_count

    def load_from_xml(self, xml_filename, store_tool_path=True) -> None:
        try:
            tree = util.parse_xml(xml_filename)
        except OSError as e:
            if e.errno != errno.ENOENT or self.app.config.is_set("data_manager_config_file"):
                raise
            return  # default config option and it doesn't exist, which is fine
        except Exception as e:
            log.error(f'There was an error parsing your Data Manager config file "{xml_filename}": {e}')
            return  # we are not able to load any data managers
        root = tree.getroot()
        if root.tag != "data_managers":
            log.error(
                f'A data managers configuration must have a "data_managers" tag as the root. "{root.tag}" is present'
            )
            return
        if store_tool_path:
            tool_path = root.get("tool_path", None)
            if tool_path is None:
                tool_path = self.app.config.tool_path
            if not tool_path:
                tool_path = "."
            self.tool_path = tool_path
        for data_manager_elem in root.findall("data_manager"):
            if not self.load_manager_from_elem(data_manager_elem, tool_path=self.tool_path):
                # Wasn't able to load manager, could happen when galaxy is managed by planemo.
                # Fall back to loading relative to the data_manager_conf.xml file
                tool_path = os.path.dirname(xml_filename)
                self.load_manager_from_elem(data_manager_elem, tool_path=tool_path)

    def load_manager_from_elem(self, data_manager_elem, tool_path=None, add_manager=True) -> Optional["DataManager"]:
        try:
            data_manager = DataManager(self, data_manager_elem, tool_path=tool_path)
        except OSError as e:
            if e.errno == errno.ENOENT:
                # File does not exist
                return None
        except Exception:
            log.exception("Error loading data_manager")
            return None
        if add_manager:
            self.add_manager(data_manager)
        log.debug(f"Loaded Data Manager: {data_manager.id}")
        return data_manager

    def add_manager(self, data_manager):
        if data_manager.id in self.data_managers:
            log.warning(f"A data manager has been defined twice: {data_manager.id} ")
        self.data_managers[data_manager.id] = data_manager
        for data_table_name in data_manager.data_table_names:
            if data_table_name not in self.managed_data_tables:
                self.managed_data_tables[data_table_name] = []
            self.managed_data_tables[data_table_name].append(data_manager)

    def get_manager(self, *args, **kwds):
        return self.data_managers.get(*args, **kwds)

    def remove_manager(self, manager_ids: Union[str, List[str]]) -> None:
        if not isinstance(manager_ids, list):
            manager_ids = [manager_ids]
        for manager_id in manager_ids:
            data_manager = self.get_manager(manager_id, None)
            if data_manager is not None:
                del self.data_managers[manager_id]
                # remove tool from toolbox
                if data_manager.tool:
                    self.app.toolbox.remove_tool_by_id(data_manager.tool.id)
                # determine if any data_tables are no longer tracked
                for data_table_name in data_manager.data_table_names:
                    remove_data_table_tracking = True
                    for other_data_manager in self.data_managers.values():
                        if data_table_name in other_data_manager.data_table_names:
                            remove_data_table_tracking = False
                            break
                    if remove_data_table_tracking and data_table_name in self.managed_data_tables:
                        del self.managed_data_tables[data_table_name]


class Tool(Protocol):
    name: str
    description: str
    version: str


class DataManager:
    GUID_TYPE = "data_manager"
    DEFAULT_VERSION = "0.0.1"

    tool: Optional[Tool]

    def __init__(self, data_managers: DataManagers, elem: Optional[Element] = None, tool_path: Optional[str] = None):
        self.data_managers = data_managers
        self.declared_id: Optional[str] = None
        self.name: Optional[str] = None
        self.description: Optional[str] = None
        self.version = self.DEFAULT_VERSION
        self.guid: Optional[str] = None
        self.tool = None
        self.tool_shed_repository_info: Optional[RepoInfo] = None
        self.undeclared_tables = False
        if elem is not None:
            self._load_from_element(elem, tool_path or self.data_managers.tool_path)

    def _load_from_element(self, elem: Element, tool_path: Optional[str]) -> None:
        assert (
            elem.tag == "data_manager"
        ), f'A data manager configuration must have a "data_manager" tag as the root. "{elem.tag}" is present'
        self.declared_id = elem.get("id")
        self.guid = elem.get("guid")
        path = elem.get("tool_file")
        tool_shed_repository = None
        tool_guid = None

        if path is None:
            tool_elem = elem.find("tool")
            assert (
                tool_elem is not None
            ), f"Error loading tool for data manager. Make sure that a tool_file attribute or a tool tag set has been defined:\n{util.xml_to_string(elem)}"
            path = tool_elem.get("file")
            tool_guid = tool_elem.get("guid")
            # need to determine repository info so that dependencies will work correctly
            tool_shed_repository = self.data_managers.app.toolbox.get_tool_repository_from_xml_item(tool_elem, path)
            self.tool_shed_repository_info = RepoInfo(
                tool_shed=tool_shed_repository.tool_shed,
                name=tool_shed_repository.name,
                owner=tool_shed_repository.owner,
                installed_changeset_revision=tool_shed_repository.installed_changeset_revision,
            )
            # use shed_conf_file to determine tool_path
            shed_conf_file = elem.get("shed_conf_file")
            if shed_conf_file:
                shed_conf = self.data_managers.app.toolbox.get_shed_config_dict_by_filename(shed_conf_file)
                if shed_conf:
                    tool_path = shed_conf.get("tool_path", tool_path)
        assert path is not None, f"A tool file path could not be determined:\n{util.xml_to_string(elem)}"
        assert tool_path, "A tool root path is required"
        self._load_tool(
            os.path.join(tool_path, path),
            guid=tool_guid,
            data_manager_id=self.id,
            tool_shed_repository=tool_shed_repository,
        )
        assert self.tool
        self.name = elem.get("name", self.tool.name)
        self.description = elem.get("description", self.tool.description)
        self.version = elem.get("version", self.tool.version)
        self.processor_description = convert_data_tables_xml(elem)

    @property
    def id(self):
        return self.guid or self.declared_id  # if we have a guid, we will use that as the data_manager id

    @property
    def data_table_names(self):
        return self.processor_description.data_table_names

    def _load_tool(
        self, tool_filename, guid=None, data_manager_id=None, tool_shed_repository_id=None, tool_shed_repository=None
    ):
        toolbox = self.data_managers.app.toolbox
        tool = toolbox.load_hidden_tool(
            tool_filename,
            guid=guid,
            data_manager_id=data_manager_id,
            repository_id=tool_shed_repository_id,
            tool_shed_repository=tool_shed_repository,
            use_cached=True,
        )
        self.data_managers.app.toolbox.data_manager_tools[tool.id] = tool
        self.tool = tool
        return tool

    def process_result(self, out_data: Dict[str, OutputDataset]) -> None:
        tool_data_tables = self.data_managers.app.tool_data_tables
        options = BundleProcessingOptions(
            what=f"data manager '{self.id}'",
            data_manager_path=self._data_manager_path,
            target_config_file=self.data_managers.filename,
        )
        updated_data_tables = tool_data_tables.process_bundle(
            out_data,
            self.processor_description,
            self.repo_info,
            options,
        )
        for data_table_name in updated_data_tables:
            self._reload(data_table_name)

    def write_bundle(
        self,
        out_data: Dict[str, OutputDataset],
    ) -> Dict[str, OutputDataset]:
        tool_data_tables = self.data_managers.app.tool_data_tables
        return tool_data_tables.write_bundle(
            out_data,
            self.processor_description,
            self.repo_info,
        )

    @property
    def _data_manager_path(self) -> str:
        return self.data_managers.app.config.galaxy_data_manager_data_path

    def _reload(self, data_table_name: str) -> None:
        self.data_managers.app.queue_worker.send_control_task(
            "reload_tool_data_tables", noop_self=True, kwargs={"table_name": data_table_name}
        )

    @property
    def repo_info(self) -> Optional[RepoInfo]:
        return self.tool_shed_repository_info

    # legacy stuff because tool shed code calls this...
    # data manager manual integration test provides coverage
    def get_tool_shed_repository_info_dict(self) -> Optional[dict]:
        repo_info = self.repo_info
        return repo_info.model_dump(mode="json") if repo_info else None



from sqlalchemy import select
from sqlalchemy.orm import exc as sqlalchemy_exceptions

from galaxy.exceptions import (
    InconsistentDatabase,
    InternalServerError,
    RequestParameterInvalidException,
)
from galaxy.managers import base
from galaxy.managers.context import ProvidesUserContext
from galaxy.model import (
    FormDefinition,
    FormDefinitionCurrent,
)
from galaxy.model.base import transaction
from galaxy.util import unicodify


def get_form_definitions(session):
    stmt = select(FormDefinition)
    return session.scalars(stmt)


def get_form_definitions_current(session):
    stmt = select(FormDefinitionCurrent)
    return session.scalars(stmt)


def get_filtered_form_definitions_current(session, filter):
    stmt = select(FormDefinitionCurrent).filter_by(**filter)
    return session.scalars(stmt)


def get_form(trans, form_id):
    """Get a FormDefinition from the database by id."""
    form = trans.sa_session.query(FormDefinitionCurrent).get(trans.security.decode_id(form_id))
    if not form:
        return trans.show_error_message(f"Form not found for id ({str(form_id)})")
    return form


class FormManager(base.ModelManager[FormDefinitionCurrent]):
    """
    Business logic for forms.
    """

    def get(self, trans: ProvidesUserContext, form_id: int) -> FormDefinitionCurrent:
        """
        Method loads the form from the DB based on the given form id.

        :param  form_id:      id of the form to load from the DB
        :type   form_id:      int

        :returns:   the loaded Form object
        :rtype:     galaxy.model.FormDefinitionCurrent

        :raises: InconsistentDatabase, RequestParameterInvalidException, InternalServerError
        """
        try:
            stmt = select(FormDefinitionCurrent).where(FormDefinitionCurrent.id == form_id)
            form = self.session().execute(stmt).scalar_one()
        except sqlalchemy_exceptions.MultipleResultsFound:
            raise InconsistentDatabase("Multiple forms found with the same id.")
        except sqlalchemy_exceptions.NoResultFound:
            raise RequestParameterInvalidException("No accessible form found with the id provided.")
        except Exception as e:
            raise InternalServerError(f"Error loading from the database.{unicodify(e)}")
        return form

    def delete(self, trans: ProvidesUserContext, form: FormDefinitionCurrent) -> FormDefinitionCurrent:
        form.deleted = True
        trans.sa_session.add(form)
        with transaction(trans.sa_session):
            trans.sa_session.commit()
        return form

    def undelete(self, trans: ProvidesUserContext, form: FormDefinitionCurrent) -> FormDefinitionCurrent:
        form.deleted = False
        trans.sa_session.add(form)
        with transaction(trans.sa_session):
            trans.sa_session.commit()
        return form



from __future__ import print_function

import sys


def stop_err(msg):
    sys.stderr.write(msg)
    sys.exit()


def __main__():
    try:
        infile = open(sys.argv[1])
        outfile = open(sys.argv[2], "w")
    except Exception:
        stop_err("Cannot open or create a file\n")

    if len(sys.argv) < 4:
        stop_err("No columns to merge")
    else:
        cols = sys.argv[3:]

    skipped_lines = 0

    for line in infile:
        line = line.rstrip("\r\n")
        if line and not line.startswith("#"):
            fields = line.split("\t")
            line += "\t"
            for col in cols:
                try:
                    line += fields[int(col) - 1]
                except Exception:
                    skipped_lines += 1

            print(line, file=outfile)

    if skipped_lines > 0:
        print("Skipped %d invalid lines" % skipped_lines)


if __name__ == "__main__":
    __main__()



from datetime import datetime
from typing import (
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
)

from pydantic import ValidationError
from sqlalchemy import (
    and_,
    delete,
    false,
    func,
    or_,
    select,
    union,
    update,
)
from sqlalchemy.sql import Select
from typing_extensions import Protocol

from galaxy.config import GalaxyAppConfiguration
from galaxy.exceptions import (
    ConfigDoesNotAllowException,
    ObjectNotFound,
)
from galaxy.model import (
    GroupRoleAssociation,
    Notification,
    User,
    UserGroupAssociation,
    UserNotificationAssociation,
    UserRoleAssociation,
)
from galaxy.model.base import transaction
from galaxy.model.scoped_session import galaxy_scoped_session
from galaxy.schema.notifications import (
    BroadcastNotificationCreateRequest,
    MandatoryNotificationCategory,
    NotificationBroadcastUpdateRequest,
    NotificationCreateData,
    NotificationCreateRequest,
    NotificationRecipients,
    PersonalNotificationCategory,
    UpdateUserNotificationPreferencesRequest,
    UserNotificationPreferences,
    UserNotificationUpdateRequest,
)

NOTIFICATION_PREFERENCES_SECTION_NAME = "notifications"


class CleanupResultSummary(NamedTuple):
    deleted_notifications_count: int
    deleted_associations_count: int


class NotificationManager:
    """Manager class to interact with the database models related with Notifications."""

    def __init__(self, sa_session: galaxy_scoped_session, config: GalaxyAppConfiguration):
        self.sa_session = sa_session
        self.config = config
        self.recipient_resolver = NotificationRecipientResolver(strategy=DefaultStrategy(sa_session))
        self.user_notification_columns = [
            Notification.id,
            Notification.source,
            Notification.category,
            Notification.variant,
            Notification.create_time,
            UserNotificationAssociation.update_time,
            Notification.publication_time,
            Notification.expiration_time,
            Notification.content,
            UserNotificationAssociation.seen_time,
            UserNotificationAssociation.deleted,
        ]
        self.broadcast_notification_columns = [
            Notification.id,
            Notification.source,
            Notification.category,
            Notification.variant,
            Notification.create_time,
            Notification.update_time,
            Notification.publication_time,
            Notification.expiration_time,
            Notification.content,
        ]

    @property
    def notifications_enabled(self):
        return self.config.enable_notification_system

    @property
    def _now(self):
        return datetime.utcnow()

    @property
    def _notification_is_active(self):
        """Condition to filter active (published and not expired) notifications only."""
        return and_(
            Notification.publication_time <= self._now,
            Notification.expiration_time > self._now,
        )

    def ensure_notifications_enabled(self):
        if not self.notifications_enabled:
            raise ConfigDoesNotAllowException("Notifications are disabled in this Galaxy.")

    def send_notification_to_recipients(self, request: NotificationCreateRequest) -> Tuple[Optional[Notification], int]:
        """
        Creates a new notification and associates it with all the recipient users.
        """
        self.ensure_notifications_enabled()
        recipient_users = self.recipient_resolver.resolve(request.recipients)
        notifications_sent = len(recipient_users)
        notification = self._create_notification_model(request.notification)
        self.sa_session.add(notification)
        self._send_to_users(notification, recipient_users)
        with transaction(self.sa_session):
            self.sa_session.commit()
        return notification, notifications_sent

    def _send_to_users(self, notification: Notification, users: List[User]):
        # TODO: Move this potentially expensive operation to a task?
        for user in users:
            if self._user_is_subscribed_to_notification(user, notification.category):
                user_notification_association = UserNotificationAssociation(user, notification)
                self.sa_session.add(user_notification_association)

    def _user_is_subscribed_to_notification(self, user: User, category: PersonalNotificationCategory) -> bool:
        notification_preferences = self.get_user_notification_preferences(user)
        category_settings = notification_preferences.get(category)
        return category_settings.enabled

    def create_broadcast_notification(self, request: BroadcastNotificationCreateRequest):
        """Creates a broadcasted notification.

        This kind of notification is not explicitly associated with any specific user but it is accessible by all users.
        """
        self.ensure_notifications_enabled()
        notification = self._create_notification_model(request)
        self.sa_session.add(notification)
        with transaction(self.sa_session):
            self.sa_session.commit()
        return notification

    def get_user_notification(self, user: User, notification_id: int, active_only: Optional[bool] = True):
        """
        Displays a notification belonging to the user.
        """
        stmt = self._user_notifications_query(user, active_only=active_only)
        stmt = stmt.where(Notification.id == notification_id)

        result = self.sa_session.execute(stmt).fetchone()
        if result is None:
            raise ObjectNotFound
        return result

    def get_user_notifications(
        self,
        user: User,
        limit: Optional[int] = 50,
        offset: Optional[int] = None,
        since: Optional[datetime] = None,
    ):
        """
        Displays the list of notifications belonging to the user.
        """
        stmt = self._user_notifications_query(user, since)
        if offset:
            stmt = stmt.offset(offset)
        if limit:
            stmt = stmt.limit(limit)

        result = self.sa_session.execute(stmt).fetchall()
        return result

    def get_user_total_unread_notification_count(self, user: User) -> int:
        """
        Returns the total number of unread notifications of the user.
        Only published and not expired notifications are accounted.
        """
        stmt = (
            select(func.count(UserNotificationAssociation.id))
            .select_from(UserNotificationAssociation)
            .join(
                Notification,
                Notification.id == UserNotificationAssociation.notification_id,
            )
            .where(self._notification_is_active)
            .where(
                and_(
                    UserNotificationAssociation.user_id == user.id,
                    UserNotificationAssociation.deleted == false(),
                    UserNotificationAssociation.seen_time.is_(None),
                )
            )
        )
        result = self.sa_session.execute(stmt).scalar()
        return result

    def get_broadcasted_notification(self, notification_id: int, active_only: Optional[bool] = True):
        stmt = (
            select(*self.broadcast_notification_columns)
            .select_from(Notification)
            .where(
                and_(
                    Notification.id == notification_id,
                    Notification.category == MandatoryNotificationCategory.broadcast,
                )
            )
        )
        if active_only:
            stmt = stmt.where(self._notification_is_active)
        result = self.sa_session.execute(stmt).fetchone()
        if result is None:
            raise ObjectNotFound
        return result

    def get_all_broadcasted_notifications(self, since: Optional[datetime] = None, active_only: Optional[bool] = True):
        stmt = self._broadcasted_notifications_query(since, active_only)
        result = self.sa_session.execute(stmt).fetchall()
        return result

    def update_user_notifications(
        self, user: User, notification_ids: Set[int], request: UserNotificationUpdateRequest
    ) -> int:
        """Updates a batch of notifications associated with the user using the requested values."""
        updated_row_count = 0
        stmt = update(UserNotificationAssociation).where(
            and_(
                UserNotificationAssociation.user_id == user.id,
                UserNotificationAssociation.notification_id.in_(notification_ids),
            )
        )
        if request.seen is not None:
            seen_time = self._now if request.seen else None
            stmt = stmt.values(seen_time=seen_time)
        if request.deleted is not None:
            stmt = stmt.values(deleted=request.deleted)
        result = self.sa_session.execute(stmt)
        updated_row_count = result.rowcount
        with transaction(self.sa_session):
            self.sa_session.commit()
        return updated_row_count

    def update_broadcasted_notification(self, notification_id: int, request: NotificationBroadcastUpdateRequest) -> int:
        """Updates a single broadcasted notification with the requested values."""
        updated_row_count = 0
        stmt = update(Notification).where(
            and_(
                Notification.id == notification_id,
                Notification.category == MandatoryNotificationCategory.broadcast,
            )
        )
        if request.source is not None:
            stmt = stmt.values(source=request.source)
        if request.variant is not None:
            stmt = stmt.values(variant=request.variant)
        if request.publication_time is not None:
            stmt = stmt.values(publication_time=request.publication_time)
        if request.expiration_time is not None:
            stmt = stmt.values(expiration_time=request.expiration_time)
        if request.content is not None:
            stmt = stmt.values(content=request.content.json())
        result = self.sa_session.execute(stmt)
        updated_row_count = result.rowcount
        with transaction(self.sa_session):
            self.sa_session.commit()
        return updated_row_count

    def get_user_notification_preferences(self, user: User) -> UserNotificationPreferences:
        """Gets the user's current notification preferences or the default ones if no preferences are set."""
        current_notification_preferences = (
            user.preferences[NOTIFICATION_PREFERENCES_SECTION_NAME]
            if NOTIFICATION_PREFERENCES_SECTION_NAME in user.preferences
            else None
        )
        try:
            return UserNotificationPreferences.parse_raw(current_notification_preferences)
        except ValidationError:
            # Gracefully return default preferences is they don't exist or get corrupted
            return UserNotificationPreferences.default()

    def update_user_notification_preferences(
        self, user: User, request: UpdateUserNotificationPreferencesRequest
    ) -> UserNotificationPreferences:
        """Updates the user's notification preferences with the requested changes."""
        notification_preferences = self.get_user_notification_preferences(user)
        notification_preferences.update(request.preferences)
        user.preferences[NOTIFICATION_PREFERENCES_SECTION_NAME] = notification_preferences.json()
        with transaction(self.sa_session):
            self.sa_session.commit()
        return notification_preferences

    def cleanup_expired_notifications(self) -> CleanupResultSummary:
        """
        Permanently removes from the database all notifications (and user associations) that have expired.
        """
        notification_has_expired = Notification.expiration_time <= self._now

        expired_notifications_stmt = select(Notification.id).where(notification_has_expired)
        delete_stmt = delete(UserNotificationAssociation).where(
            UserNotificationAssociation.notification_id.in_(expired_notifications_stmt.subquery())
        )
        result = self.sa_session.execute(delete_stmt, execution_options={"synchronize_session": False})
        deleted_associations_count = result.rowcount

        delete_stmt = delete(Notification).where(notification_has_expired)
        result = self.sa_session.execute(delete_stmt)
        deleted_notifications_count = result.rowcount

        with transaction(self.sa_session):
            self.sa_session.commit()

        return CleanupResultSummary(deleted_notifications_count, deleted_associations_count)

    def _create_notification_model(self, payload: NotificationCreateData):
        notification = Notification(
            payload.source,
            payload.category,
            payload.variant,
            payload.content.model_dump(),
        )
        notification.publication_time = payload.publication_time
        notification.expiration_time = payload.expiration_time
        return notification

    def _user_notifications_query(
        self, user: User, since: Optional[datetime] = None, active_only: Optional[bool] = True
    ):
        stmt = (
            select(*self.user_notification_columns)
            .select_from(Notification)
            .join(
                UserNotificationAssociation,
                UserNotificationAssociation.notification_id == Notification.id,
            )
            .where(UserNotificationAssociation.user_id == user.id)
        )
        if since is not None:
            stmt = stmt.where(
                or_(
                    UserNotificationAssociation.update_time >= since,
                    Notification.publication_time >= since,
                )
            )
        if active_only:
            stmt = stmt.where(self._notification_is_active)
            stmt = stmt.where(UserNotificationAssociation.deleted == false())

        return stmt

    def _broadcasted_notifications_query(self, since: Optional[datetime] = None, active_only: Optional[bool] = True):
        stmt = (
            select(*self.broadcast_notification_columns)
            .select_from(Notification)
            .where(Notification.category == MandatoryNotificationCategory.broadcast)
        )
        if since is not None:
            stmt = stmt.where(
                or_(
                    Notification.update_time >= since,
                    Notification.publication_time >= since,
                )
            )
        if active_only:
            stmt = stmt.where(self._notification_is_active)
        return stmt


class NotificationRecipientResolverStrategy(Protocol):
    def resolve_users(self, recipients: NotificationRecipients) -> List[User]:
        pass


class NotificationRecipientResolver:
    """Resolves a set of NotificationRecipients to a list of unique users using a specific strategy."""

    def __init__(self, strategy: NotificationRecipientResolverStrategy):
        self.strategy = strategy

    def resolve(self, recipients: NotificationRecipients) -> List[User]:
        """Given individual user, group and roles ids as recipients, obtains the unique list of users.

        The resulting list will contain only unique users even if the same user id might have been provided more
        than once as `user_ids` input or implicitly in groups or roles.
        """
        return self.strategy.resolve_users(recipients)


class DefaultStrategy(NotificationRecipientResolverStrategy):
    """Resolves the recipient users using an iterative approach."""

    def __init__(self, sa_session: galaxy_scoped_session):
        self.sa_session = sa_session

    def resolve_users(self, recipients: NotificationRecipients) -> List[User]:
        unique_user_ids: Set[int] = set(recipients.user_ids)

        all_group_ids, all_role_ids = self._expand_group_and_roles_ids(
            set(recipients.group_ids), set(recipients.role_ids)
        )

        user_ids_from_groups_stmt = self._get_all_user_ids_from_groups_query(all_group_ids)
        user_ids_from_roles_stmt = self._get_all_user_ids_from_roles_query(all_role_ids)

        union_stmt = union(user_ids_from_groups_stmt, user_ids_from_roles_stmt)
        user_ids_from_groups_and_roles = {id for id, in self.sa_session.execute(union_stmt)}
        unique_user_ids.update(user_ids_from_groups_and_roles)

        stmt = select(User).where(User.id.in_(unique_user_ids))
        return self.sa_session.scalars(stmt).all()

    def _get_all_user_ids_from_roles_query(self, role_ids: Set[int]) -> Select:
        stmt = (
            select(UserRoleAssociation.user_id)
            .select_from(UserRoleAssociation)
            .where(UserRoleAssociation.role_id.in_(role_ids))
            .distinct()
        )
        return stmt

    def _get_all_user_ids_from_groups_query(self, group_ids: Set[int]) -> Select:
        stmt = (
            select(UserGroupAssociation.user_id)
            .select_from(UserGroupAssociation)
            .where(UserGroupAssociation.group_id.in_(group_ids))
            .distinct()
        )
        return stmt

    def _expand_group_and_roles_ids(self, group_ids: Set[int], role_ids: Set[int]) -> Tuple[Set[int], Set[int]]:
        """Given a set of group and roles IDs, it expands those sets (non-recursively) by including sub-groups or sub-roles
        indirectly associated with them.
        """
        processed_group_ids: Set[int] = set()
        processed_role_ids: Set[int] = set()

        while True:
            # Get group IDs associated with any of the given role IDs
            stmt = (
                select(GroupRoleAssociation.group_id)
                .select_from(GroupRoleAssociation)
                .where(GroupRoleAssociation.role_id.in_(role_ids))
                .distinct()
            )
            group_ids_from_roles = {id for id, in self.sa_session.execute(stmt) if id is not None}
            new_group_ids = group_ids_from_roles - processed_group_ids

            # Get role IDs associated with any of the given group IDs
            stmt = (
                select(GroupRoleAssociation.role_id)
                .select_from(GroupRoleAssociation)
                .where(GroupRoleAssociation.group_id.in_(group_ids))
                .distinct()
            )
            role_ids_from_groups = {id for id, in self.sa_session.execute(stmt) if id is not None}
            new_role_ids = role_ids_from_groups - processed_role_ids

            # Stop if there are no new group or role IDs to process
            if not new_group_ids and not new_role_ids:
                break

            # Add new group and role IDs to the respective sets
            group_ids.update(new_group_ids)
            role_ids.update(new_role_ids)

            # Add new group and role IDs to the processed sets
            processed_group_ids.update(new_group_ids)
            processed_role_ids.update(new_role_ids)

        return group_ids, role_ids


class RecursiveCTEStrategy(NotificationRecipientResolverStrategy):
    def resolve_users(self, recipients: NotificationRecipients) -> List[User]:
        # TODO Implement resolver using recursive CTEs?
        return []



#!/usr/bin/env python
# Guruprasad Ananda
"""
Converts coordinates from one build/assembly to another using liftOver binary and mapping files downloaded from UCSC.
"""

import os
import re
import subprocess
import sys
import tempfile
from typing import Union


def stop_err(msg):
    sys.stderr.write(msg)
    sys.exit()


def safe_bed_file(infile):
    """Make a BED file with track and browser lines ready for liftOver.

    liftOver will fail with track or browser lines. We can make it happy
    by converting these to comments. See:

    https://lists.soe.ucsc.edu/pipermail/genome/2007-May/013561.html
    """
    fix_pat = re.compile("^(track|browser)")
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as out_handle, open(infile, "r") as in_handle:
        for line in in_handle:
            if fix_pat.match(line):
                line = "#" + line
            out_handle.write(line)
    return out_handle.name


if len(sys.argv) < 9:
    stop_err(
        "USAGE: prog input out_file1 out_file2 input_dbkey output_dbkey infile_type minMatch multiple <minChainT> <minChainQ> <minSizeQ>"
    )

infile = sys.argv[1]
outfile1 = sys.argv[2]
outfile2 = sys.argv[3]
in_dbkey = sys.argv[4]
mapfilepath = sys.argv[5]
infile_type = sys.argv[6]
gff_option = ""
if infile_type == "gff":
    gff_option = "-gff "
minMatch: Union[str, float] = sys.argv[7]
multiple = int(sys.argv[8])
multiple_option = ""
if multiple:
    minChainT = sys.argv[9]
    minChainQ = sys.argv[10]
    minSizeQ = sys.argv[11]
    multiple_option = " -multiple -minChainT=%s -minChainQ=%s -minSizeQ=%s " % (minChainT, minChainQ, minSizeQ)

try:
    assert float(minMatch)
except Exception:
    minMatch = 0.1
# ensure dbkey is set
if in_dbkey == "?":
    stop_err("Input dataset genome build unspecified, click the pencil icon in the history item to specify it.")

if not os.path.isfile(mapfilepath):
    stop_err("%s mapping is not currently available." % (mapfilepath.split("/")[-1].split(".")[0]))

safe_infile = safe_bed_file(infile)
cmd_line = (
    "liftOver "
    + gff_option
    + "-minMatch="
    + str(minMatch)
    + multiple_option
    + " "
    + safe_infile
    + " "
    + mapfilepath
    + " "
    + outfile1
    + " "
    + outfile2
    + "  > /dev/null"
)

try:
    # have to nest try-except in try-finally to handle 2.4
    try:
        proc = subprocess.Popen(args=cmd_line, shell=True, stderr=subprocess.PIPE)
        stderr = proc.communicate()[1]
        if proc.returncode != 0:
            raise Exception(stderr)
    except Exception as e:
        raise Exception("Exception caught attempting conversion: " + str(e))
finally:
    os.remove(safe_infile)



import logging
import os
import re
import shutil
from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)
from urllib.error import HTTPError

from markupsafe import escape
from sqlalchemy import (
    and_,
    false,
    or_,
)
from sqlalchemy.orm import joinedload

from galaxy import util
from galaxy.model.base import transaction
from galaxy.model.scoped_session import install_model_scoped_session
from galaxy.model.tool_shed_install import ToolShedRepository
from galaxy.tool_shed.util import basic_util
from galaxy.util.tool_shed import (
    common_util,
    encoding_util,
)
from galaxy.util.tool_shed.tool_shed_registry import Registry

log = logging.getLogger(__name__)

VALID_REPOSITORYNAME_RE = re.compile(r"^[a-z0-9\_]+$")


def check_for_updates(
    tool_shed_registry: Registry,
    install_model_context: install_model_scoped_session,
    repository_id: Optional[int] = None,
) -> Tuple[str, str]:
    message = ""
    status = "ok"
    if repository_id is None:
        success_count = 0
        repository_names_not_updated = []
        updated_count = 0
        for repository in install_model_context.query(ToolShedRepository).filter(
            ToolShedRepository.table.c.deleted == false()
        ):
            ok, updated = _check_or_update_tool_shed_status_for_installed_repository(
                tool_shed_registry, install_model_context, repository
            )
            if ok:
                success_count += 1
            else:
                repository_names_not_updated.append(f"<b>{escape(str(repository.name))}</b>")
            if updated:
                updated_count += 1
        message = "Checked the status in the tool shed for %d repositories.  " % success_count
        message += "Updated the tool shed status for %d repositories.  " % updated_count
        if repository_names_not_updated:
            message += "Unable to retrieve status from the tool shed for the following repositories:\n"
            message += ", ".join(repository_names_not_updated)
    else:
        repository = install_model_context.get(ToolShedRepository, repository_id)
        ok, updated = _check_or_update_tool_shed_status_for_installed_repository(
            tool_shed_registry, install_model_context, repository
        )
        if ok:
            if updated:
                message = f"The tool shed status for repository <b>{escape(str(repository.name))}</b> has been updated."
            else:
                message = (
                    f"The status has not changed in the tool shed for repository <b>{escape(str(repository.name))}</b>."
                )
        else:
            message = (
                f"Unable to retrieve status from the tool shed for repository <b>{escape(str(repository.name))}</b>."
            )
            status = "error"
    return message, status


def _check_or_update_tool_shed_status_for_installed_repository(
    tool_shed_registry: Registry, install_model_context: install_model_scoped_session, repository: ToolShedRepository
) -> Tuple[bool, bool]:
    updated = False
    tool_shed_status_dict = get_tool_shed_status_for(tool_shed_registry, repository)
    if tool_shed_status_dict:
        ok = True
        if tool_shed_status_dict != repository.tool_shed_status:
            repository.tool_shed_status = tool_shed_status_dict
            session = install_model_context
            session.add(repository)
            with transaction(session):
                session.commit()

            updated = True
    else:
        ok = False
    return ok, updated


def create_or_update_tool_shed_repository(
    app,
    name,
    description,
    installed_changeset_revision,
    ctx_rev,
    repository_clone_url,
    status,
    metadata_dict=None,
    current_changeset_revision=None,
    owner="",
    dist_to_shed=False,
):
    """
    Update a tool shed repository record in the Galaxy database with the new information received.
    If a record defined by the received tool shed, repository name and owner does not exist, create
    a new record with the received information.
    """
    metadata_dict = metadata_dict or {}
    # The received value for dist_to_shed will be True if the ToolMigrationManager is installing a repository
    # that contains tools or datatypes that used to be in the Galaxy distribution, but have been moved
    # to the main Galaxy tool shed.
    if current_changeset_revision is None:
        # The current_changeset_revision is not passed if a repository is being installed for the first
        # time.  If a previously installed repository was later uninstalled, this value should be received
        # as the value of that change set to which the repository had been updated just prior to it being
        # uninstalled.
        current_changeset_revision = installed_changeset_revision
    context = app.install_model.context
    tool_shed = get_tool_shed_from_clone_url(repository_clone_url)
    if not owner:
        owner = get_repository_owner_from_clone_url(repository_clone_url)
    if status in [app.install_model.ToolShedRepository.installation_status.DEACTIVATED]:
        deleted = True
        uninstalled = False
    elif status in [app.install_model.ToolShedRepository.installation_status.UNINSTALLED]:
        deleted = True
        uninstalled = True
    else:
        deleted = False
        uninstalled = False
    tool_shed_repository = get_installed_repository(
        app, tool_shed=tool_shed, name=name, owner=owner, installed_changeset_revision=installed_changeset_revision
    )
    if tool_shed_repository:
        log.debug(
            "Updating an existing row for repository '%s' in the tool_shed_repository table, status set to '%s'.",
            name,
            status,
        )
        tool_shed_repository.description = description
        tool_shed_repository.changeset_revision = current_changeset_revision
        tool_shed_repository.ctx_rev = ctx_rev
        tool_shed_repository.metadata_ = metadata_dict
        tool_shed_repository.deleted = deleted
        tool_shed_repository.uninstalled = uninstalled
        tool_shed_repository.status = status
    else:
        log.debug(
            "Adding new row for repository '%s' in the tool_shed_repository table, status set to '%s'.", name, status
        )
        tool_shed_repository = app.install_model.ToolShedRepository(
            tool_shed=tool_shed,
            name=name,
            description=description,
            owner=owner,
            installed_changeset_revision=installed_changeset_revision,
            changeset_revision=current_changeset_revision,
            ctx_rev=ctx_rev,
            metadata_=metadata_dict,
            dist_to_shed=dist_to_shed,
            deleted=deleted,
            uninstalled=uninstalled,
            status=status,
        )
    context.add(tool_shed_repository)
    with transaction(context):
        context.commit()
    return tool_shed_repository


def extract_components_from_tuple(repository_components_tuple):
    """Extract the repository components from the provided tuple in a backward-compatible manner."""
    toolshed = repository_components_tuple[0]
    name = repository_components_tuple[1]
    owner = repository_components_tuple[2]
    changeset_revision = repository_components_tuple[3]
    components_list = [toolshed, name, owner, changeset_revision]
    if len(repository_components_tuple) == 5:
        toolshed, name, owner, changeset_revision, prior_installation_required = repository_components_tuple
        components_list = [toolshed, name, owner, changeset_revision, prior_installation_required]
    elif len(repository_components_tuple) == 6:
        (
            toolshed,
            name,
            owner,
            changeset_revision,
            prior_installation_required,
            only_if_compiling_contained_td,
        ) = repository_components_tuple
        components_list = [
            toolshed,
            name,
            owner,
            changeset_revision,
            prior_installation_required,
            only_if_compiling_contained_td,
        ]
    return components_list


def generate_tool_shed_repository_install_dir(repository_clone_url, changeset_revision):
    """
    Generate a repository installation directory that guarantees repositories with the same
    name will always be installed in different directories.  The tool path will be of the form:
    <tool shed url>/repos/<repository owner>/<repository name>/<installed changeset revision>
    """
    tmp_url = common_util.remove_protocol_and_user_from_clone_url(repository_clone_url)
    # Now tmp_url is something like: bx.psu.edu:9009/repos/some_username/column
    items = tmp_url.split("/repos/")
    tool_shed_url = items[0]
    if len(items) == 1:
        raise Exception(f"Processing an invalid tool shed clone URL {repository_clone_url} - tmp_url {tmp_url}")
    repo_path = items[1]
    tool_shed_url = common_util.remove_port_from_tool_shed_url(tool_shed_url)
    return "/".join((tool_shed_url, "repos", repo_path, changeset_revision))


def get_absolute_path_to_file_in_repository(repo_files_dir, file_name):
    """Return the absolute path to a specified disk file contained in a repository."""
    stripped_file_name = basic_util.strip_path(file_name)
    file_path = None
    for root, _, files in os.walk(repo_files_dir):
        if root.find(".hg") < 0:
            for name in files:
                if name == stripped_file_name:
                    return os.path.abspath(os.path.join(root, name))
    return file_path


def get_ids_of_tool_shed_repositories_being_installed(app, as_string=False):
    installing_repository_ids = []
    new_status = app.install_model.ToolShedRepository.installation_status.NEW
    cloning_status = app.install_model.ToolShedRepository.installation_status.CLONING
    setting_tool_versions_status = app.install_model.ToolShedRepository.installation_status.SETTING_TOOL_VERSIONS
    installing_dependencies_status = (
        app.install_model.ToolShedRepository.installation_status.INSTALLING_TOOL_DEPENDENCIES
    )
    loading_datatypes_status = app.install_model.ToolShedRepository.installation_status.LOADING_PROPRIETARY_DATATYPES
    for tool_shed_repository in app.install_model.context.query(app.install_model.ToolShedRepository).filter(
        or_(
            app.install_model.ToolShedRepository.status == new_status,
            app.install_model.ToolShedRepository.status == cloning_status,
            app.install_model.ToolShedRepository.status == setting_tool_versions_status,
            app.install_model.ToolShedRepository.status == installing_dependencies_status,
            app.install_model.ToolShedRepository.status == loading_datatypes_status,
        )
    ):
        installing_repository_ids.append(app.security.encode_id(tool_shed_repository.id))
    if as_string:
        return ",".join(installing_repository_ids)
    return installing_repository_ids


def get_installed_repository(
    app,
    tool_shed=None,
    name=None,
    owner=None,
    changeset_revision=None,
    installed_changeset_revision=None,
    repository_id=None,
    from_cache=False,
):
    """
    Return a tool shed repository database record defined by the combination of a toolshed, repository name,
    repository owner and either current or originally installed changeset_revision.
    """
    # We store the port, if one exists, in the database.
    tool_shed = common_util.remove_protocol_from_tool_shed_url(tool_shed)
    if from_cache:
        tsr_cache = getattr(app, "tool_shed_repository_cache", None)
        if tsr_cache:
            return app.tool_shed_repository_cache.get_installed_repository(
                tool_shed=tool_shed,
                name=name,
                owner=owner,
                installed_changeset_revision=installed_changeset_revision,
                changeset_revision=changeset_revision,
                repository_id=repository_id,
            )
    query = app.install_model.context.query(app.install_model.ToolShedRepository)
    if repository_id:
        clause_list = [app.install_model.ToolShedRepository.table.c.id == repository_id]
    else:
        clause_list = [
            app.install_model.ToolShedRepository.table.c.tool_shed == tool_shed,
            app.install_model.ToolShedRepository.table.c.name == name,
            app.install_model.ToolShedRepository.table.c.owner == owner,
        ]
    if changeset_revision is not None:
        clause_list.append(app.install_model.ToolShedRepository.table.c.changeset_revision == changeset_revision)
    if installed_changeset_revision is not None:
        clause_list.append(
            app.install_model.ToolShedRepository.table.c.installed_changeset_revision == installed_changeset_revision
        )
    return query.filter(and_(*clause_list)).first()


def get_installed_tool_shed_repository(app, id):
    """Get a tool shed repository record from the Galaxy database defined by the id."""
    rval = []
    if isinstance(id, list):
        return_list = True
    else:
        id = [id]
        return_list = False
    repository_ids = [app.security.decode_id(i) for i in id]
    rval = [get_installed_repository(app=app, repository_id=repo_id, from_cache=False) for repo_id in repository_ids]
    if return_list:
        return rval
    return rval[0]


def get_prior_import_or_install_required_dict(app, tsr_ids, repo_info_dicts):
    """
    This method is used in the Tool Shed when exporting a repository and its dependencies,
    and in Galaxy when a repository and its dependencies are being installed.  Return a
    dictionary whose keys are the received tsr_ids and whose values are a list of tsr_ids,
    each of which is contained in the received list of tsr_ids and whose associated repository
    must be imported or installed prior to the repository associated with the tsr_id key.
    """
    # Initialize the dictionary.
    prior_import_or_install_required_dict = {}
    for tsr_id in tsr_ids:
        prior_import_or_install_required_dict[tsr_id] = []
    # Inspect the repository dependencies for each repository about to be installed and populate the dictionary.
    for repo_info_dict in repo_info_dicts:
        repository, repository_dependencies = get_repository_and_repository_dependencies_from_repo_info_dict(
            app, repo_info_dict
        )
        if repository:
            encoded_repository_id = app.security.encode_id(repository.id)
            if encoded_repository_id in tsr_ids:
                # We've located the database table record for one of the repositories we're about to install, so find out if it has any repository
                # dependencies that require prior installation.
                prior_import_or_install_ids = get_repository_ids_requiring_prior_import_or_install(
                    app, tsr_ids, repository_dependencies
                )
                prior_import_or_install_required_dict[encoded_repository_id] = prior_import_or_install_ids
    return prior_import_or_install_required_dict


ToolDependenciesDictT = Dict[str, Union[Dict[str, Any], List[Dict[str, Any]]]]
OldRepositoryTupleT = Tuple[str, str, str, str, str, ToolDependenciesDictT]
RepositoryTupleT = Tuple[str, str, str, str, str, Optional[Any], ToolDependenciesDictT]
AnyRepositoryTupleT = Union[OldRepositoryTupleT, RepositoryTupleT]


def get_repo_info_tuple_contents(repo_info_tuple: AnyRepositoryTupleT) -> RepositoryTupleT:
    """Take care in handling the repo_info_tuple as it evolves over time as new tool shed features are introduced."""
    if len(repo_info_tuple) == 6:
        old_repo_info = cast(OldRepositoryTupleT, repo_info_tuple)
        (
            description,
            repository_clone_url,
            changeset_revision,
            ctx_rev,
            repository_owner,
            tool_dependencies,
        ) = old_repo_info
        repository_dependencies = None
    elif len(repo_info_tuple) == 7:
        repo_info = cast(RepositoryTupleT, repo_info_tuple)
        (
            description,
            repository_clone_url,
            changeset_revision,
            ctx_rev,
            repository_owner,
            repository_dependencies,
            tool_dependencies,
        ) = repo_info
    return (
        description,
        repository_clone_url,
        changeset_revision,
        ctx_rev,
        repository_owner,
        repository_dependencies,
        tool_dependencies,
    )


def get_repository_admin_role_name(repository_name, repository_owner):
    return f"{repository_name}_{repository_owner}_admin"


def get_repository_and_repository_dependencies_from_repo_info_dict(app, repo_info_dict):
    """Return a tool_shed_repository or repository record defined by the information in the received repo_info_dict."""
    repository_name = list(repo_info_dict.keys())[0]
    repo_info_tuple = repo_info_dict[repository_name]
    (
        description,
        repository_clone_url,
        changeset_revision,
        ctx_rev,
        repository_owner,
        repository_dependencies,
        tool_dependencies,
    ) = get_repo_info_tuple_contents(repo_info_tuple)
    if hasattr(app, "install_model"):
        # In a tool shed client (Galaxy, or something install repositories like Galaxy)
        tool_shed = get_tool_shed_from_clone_url(repository_clone_url)
        repository = get_repository_for_dependency_relationship(
            app, tool_shed, repository_name, repository_owner, changeset_revision
        )
    else:
        # We're in the tool shed.
        repository = get_repository_by_name_and_owner(app, repository_name, repository_owner)
    return repository, repository_dependencies


def get_repository_by_id(app, id):
    """Get a repository from the database via id."""
    if is_tool_shed_client(app):
        return app.install_model.context.query(app.install_model.ToolShedRepository).get(app.security.decode_id(id))
    else:
        sa_session = app.model.session
        return sa_session.query(app.model.Repository).get(app.security.decode_id(id))


def get_repository_by_name_and_owner(app, name, owner, eagerload_columns=None):
    """Get a repository from the database via name and owner"""
    repository_query = get_repository_query(app)
    if is_tool_shed_client(app):
        return repository_query.filter(
            and_(
                app.install_model.ToolShedRepository.table.c.name == name,
                app.install_model.ToolShedRepository.table.c.owner == owner,
            )
        ).first()
    # We're in the tool shed.
    q = repository_query.filter(
        and_(
            app.model.Repository.table.c.name == name,
            app.model.User.table.c.username == owner,
            app.model.Repository.table.c.user_id == app.model.User.table.c.id,
        )
    )
    if eagerload_columns:
        q = q.options(joinedload(*eagerload_columns))
    return q.first()


def get_repository_by_name(app, name):
    """Get a repository from the database via name."""
    return get_repository_query(app).filter_by(name=name).first()


def get_repository_dependency_types(repository_dependencies):
    """
    Inspect the received list of repository_dependencies tuples and return boolean values
    for has_repository_dependencies and has_repository_dependencies_only_if_compiling_contained_td.
    """
    # Set has_repository_dependencies, which will be True only if at least one repository_dependency
    # is defined with the value of
    # only_if_compiling_contained_td as False.
    has_repository_dependencies = False
    for rd_tup in repository_dependencies:
        (
            tool_shed,
            name,
            owner,
            changeset_revision,
            prior_installation_required,
            only_if_compiling_contained_td,
        ) = common_util.parse_repository_dependency_tuple(rd_tup)
        if not util.asbool(only_if_compiling_contained_td):
            has_repository_dependencies = True
            break
    # Set has_repository_dependencies_only_if_compiling_contained_td, which will be True only if at
    # least one repository_dependency is defined with the value of only_if_compiling_contained_td as True.
    has_repository_dependencies_only_if_compiling_contained_td = False
    for rd_tup in repository_dependencies:
        (
            tool_shed,
            name,
            owner,
            changeset_revision,
            prior_installation_required,
            only_if_compiling_contained_td,
        ) = common_util.parse_repository_dependency_tuple(rd_tup)
        if util.asbool(only_if_compiling_contained_td):
            has_repository_dependencies_only_if_compiling_contained_td = True
            break
    return has_repository_dependencies, has_repository_dependencies_only_if_compiling_contained_td


def get_repository_for_dependency_relationship(app, tool_shed, name, owner, changeset_revision):
    """
    Return an installed tool_shed_repository database record that is defined by either the current changeset
    revision or the installed_changeset_revision.
    """
    # This method is used only in Galaxy, not the Tool Shed.  We store the port (if one exists) in the database.
    tool_shed = common_util.remove_protocol_from_tool_shed_url(tool_shed)
    if tool_shed is None or name is None or owner is None or changeset_revision is None:
        message = "Unable to retrieve the repository record from the database because one or more of the following "
        message += f"required parameters is None: tool_shed: {tool_shed}, name: {name}, owner: {owner}, changeset_revision: {changeset_revision}"
        raise Exception(message)
    repository = get_installed_repository(
        app=app, tool_shed=tool_shed, name=name, owner=owner, installed_changeset_revision=changeset_revision
    )
    if not repository:
        repository = get_installed_repository(
            app=app, tool_shed=tool_shed, name=name, owner=owner, changeset_revision=changeset_revision
        )
    if not repository:
        tool_shed_url = common_util.get_tool_shed_url_from_tool_shed_registry(app, tool_shed)
        repository_clone_url = os.path.join(tool_shed_url, "repos", owner, name)
        repo_info_tuple = (None, repository_clone_url, changeset_revision, None, owner, None, None)
        repository, pcr = repository_was_previously_installed(app, tool_shed_url, name, repo_info_tuple)
    if not repository:
        # The received changeset_revision is no longer installable, so get the next changeset_revision
        # in the repository's changelog in the tool shed that is associated with repository_metadata.
        tool_shed_url = common_util.get_tool_shed_url_from_tool_shed_registry(app, tool_shed)
        params = dict(name=name, owner=owner, changeset_revision=changeset_revision)
        pathspec = ["repository", "next_installable_changeset_revision"]
        text = util.url_get(
            tool_shed_url, auth=app.tool_shed_registry.url_auth(tool_shed_url), pathspec=pathspec, params=params
        )
        if text:
            repository = get_installed_repository(
                app=app, tool_shed=tool_shed, name=name, owner=owner, changeset_revision=text
            )
    return repository


def get_repository_ids_requiring_prior_import_or_install(app, tsr_ids, repository_dependencies):
    """
    This method is used in the Tool Shed when exporting a repository and its dependencies,
    and in Galaxy when a repository and its dependencies are being installed.  Inspect the
    received repository_dependencies and determine if the encoded id of each required
    repository is in the received tsr_ids.  If so, then determine whether that required
    repository should be imported / installed prior to its dependent repository.  Return a
    list of encoded repository ids, each of which is contained in the received list of tsr_ids,
    and whose associated repositories must be imported / installed prior to the dependent
    repository associated with the received repository_dependencies.
    """
    prior_tsr_ids = []
    if repository_dependencies:
        for key, rd_tups in repository_dependencies.items():
            if key in ["description", "root_key"]:
                continue
            for rd_tup in rd_tups:
                (
                    tool_shed,
                    name,
                    owner,
                    changeset_revision,
                    prior_installation_required,
                    only_if_compiling_contained_td,
                ) = common_util.parse_repository_dependency_tuple(rd_tup)
                # If only_if_compiling_contained_td is False, then the repository dependency
                # is not required to be installed prior to the dependent repository even if
                # prior_installation_required is True.  This is because the only meaningful
                # content of the repository dependency is its contained tool dependency, which
                # is required in order to compile the dependent repository's tool dependency.
                # In the scenario where the repository dependency is not installed prior to the
                # dependent repository's tool dependency compilation process, the tool dependency
                # compilation framework will install the repository dependency prior to compilation
                # of the dependent repository's tool dependency.
                if not util.asbool(only_if_compiling_contained_td):
                    if util.asbool(prior_installation_required):
                        if is_tool_shed_client(app):
                            # We store the port, if one exists, in the database.
                            tool_shed = common_util.remove_protocol_from_tool_shed_url(tool_shed)
                            repository = get_repository_for_dependency_relationship(
                                app, tool_shed, name, owner, changeset_revision
                            )
                        else:
                            repository = get_repository_by_name_and_owner(app, name, owner)
                        if repository:
                            encoded_repository_id = app.security.encode_id(repository.id)
                            if encoded_repository_id in tsr_ids:
                                prior_tsr_ids.append(encoded_repository_id)
    return prior_tsr_ids


def get_repository_owner(cleaned_repository_url):
    """Gvien a "cleaned" repository clone URL, return the owner of the repository."""
    items = cleaned_repository_url.split("/repos/")
    repo_path = items[1]
    if repo_path.startswith("/"):
        repo_path = repo_path.replace("/", "", 1)
    return repo_path.lstrip("/").split("/")[0]


def get_repository_owner_from_clone_url(repository_clone_url):
    """Given a repository clone URL, return the owner of the repository."""
    tmp_url = common_util.remove_protocol_and_user_from_clone_url(repository_clone_url)
    return get_repository_owner(tmp_url)


def get_repository_query(app):
    if is_tool_shed_client(app):
        query = app.install_model.context.query(app.install_model.ToolShedRepository)
    else:
        query = app.model.context.query(app.model.Repository)
    return query


def get_role_by_id(app, role_id):
    """Get a Role from the database by id."""
    sa_session = app.model.session
    return sa_session.query(app.model.Role).get(app.security.decode_id(role_id))


def get_tool_shed_from_clone_url(repository_clone_url):
    tmp_url = common_util.remove_protocol_and_user_from_clone_url(repository_clone_url)
    return tmp_url.split("/repos/")[0].rstrip("/")


def get_tool_shed_repository_by_id(app, repository_id) -> ToolShedRepository:
    """Return a tool shed repository database record defined by the id."""
    # This method is used only in Galaxy, not the tool shed.
    return app.install_model.context.get(ToolShedRepository, app.security.decode_id(repository_id))


def get_tool_shed_status_for(tool_shed_registry: Registry, repository: ToolShedRepository):
    tool_shed_url = tool_shed_registry.get_tool_shed_url(str(repository.tool_shed))
    assert tool_shed_url
    params = dict(name=repository.name, owner=repository.owner, changeset_revision=repository.changeset_revision)
    pathspec = ["repository", "status_for_installed_repository"]
    try:
        encoded_tool_shed_status_dict = util.url_get(
            tool_shed_url, auth=tool_shed_registry.url_auth(tool_shed_url), pathspec=pathspec, params=params
        )
        tool_shed_status_dict = encoding_util.tool_shed_decode(encoded_tool_shed_status_dict)
        return tool_shed_status_dict
    except HTTPError as e:
        # This should handle backward compatility to the Galaxy 12/20/12 release.  We used to only handle updates for an installed revision
        # using a boolean value.
        log.debug(
            "Error attempting to get tool shed status for installed repository %s: %s\nAttempting older 'check_for_updates' method.\n",
            repository.name,
            e,
        )
        pathspec = ["repository", "check_for_updates"]
        params["from_update_manager"] = True
        try:
            # The value of text will be 'true' or 'false', depending upon whether there is an update available for the installed revision.
            text = util.url_get(
                tool_shed_url, auth=tool_shed_registry.url_auth(tool_shed_url), pathspec=pathspec, params=params
            )
            return dict(revision_update=text)
        except Exception:
            # The required tool shed may be unavailable, so default the revision_update value to 'false'.
            return dict(revision_update="false")
    except Exception:
        log.exception("Error attempting to get tool shed status for installed repository %s", str(repository.name))
        return {}


def get_tool_shed_status_for_installed_repository(app, repository: ToolShedRepository):
    """
    Send a request to the tool shed to retrieve information about newer installable repository revisions,
    current revision updates, whether the repository revision is the latest downloadable revision, and
    whether the repository has been deprecated in the tool shed.  The received repository is a ToolShedRepository
    object from Galaxy.
    """
    tool_shed_registry = app.tool_shed_registry
    return get_tool_shed_status_for(tool_shed_registry, repository)


def is_tool_shed_client(app):
    """
    The tool shed and clients to the tool (i.e. Galaxy) require a lot
    of similar functionality in this file but with small differences. This
    method should determine if the app performing the action is the tool shed
    or a client of the tool shed.
    """
    return hasattr(app, "install_model")


def repository_was_previously_installed(app, tool_shed_url, repository_name, repo_info_tuple, from_tip=False):
    """
    Find out if a repository is already installed into Galaxy - there are several scenarios where this
    is necessary.  For example, this method will handle the case where the repository was previously
    installed using an older changeset_revsion, but later the repository was updated in the tool shed
    and now we're trying to install the latest changeset revision of the same repository instead of
    updating the one that was previously installed.  We'll look in the database instead of on disk since
    the repository may be currently uninstalled.
    """
    tool_shed_url = common_util.get_tool_shed_url_from_tool_shed_registry(app, tool_shed_url)
    (
        description,
        repository_clone_url,
        changeset_revision,
        ctx_rev,
        repository_owner,
        repository_dependencies,
        tool_dependencies,
    ) = get_repo_info_tuple_contents(repo_info_tuple)
    tool_shed = get_tool_shed_from_clone_url(repository_clone_url)
    # See if we can locate the repository using the value of changeset_revision.
    tool_shed_repository = get_installed_repository(
        app,
        tool_shed=tool_shed,
        name=repository_name,
        owner=repository_owner,
        installed_changeset_revision=changeset_revision,
    )
    if tool_shed_repository:
        return tool_shed_repository, changeset_revision
    # Get all previous changeset revisions from the tool shed for the repository back to, but excluding,
    # the previous valid changeset revision to see if it was previously installed using one of them.
    params = dict(
        name=repository_name,
        owner=repository_owner,
        changeset_revision=changeset_revision,
        from_tip=str(from_tip),
    )
    pathspec = ["repository", "previous_changeset_revisions"]
    text = util.url_get(
        tool_shed_url, auth=app.tool_shed_registry.url_auth(tool_shed_url), pathspec=pathspec, params=params
    )
    if text:
        changeset_revisions = util.listify(text)
        for previous_changeset_revision in changeset_revisions:
            tool_shed_repository = get_installed_repository(
                app,
                tool_shed=tool_shed,
                name=repository_name,
                owner=repository_owner,
                installed_changeset_revision=previous_changeset_revision,
            )
            if tool_shed_repository:
                return tool_shed_repository, previous_changeset_revision
    return None, None


def set_repository_attributes(app, repository, status, error_message, deleted, uninstalled, remove_from_disk=False):
    if remove_from_disk:
        relative_install_dir = repository.repo_path(app)
        if relative_install_dir:
            clone_dir = os.path.abspath(relative_install_dir)
            try:
                shutil.rmtree(clone_dir)
                log.debug("Removed repository installation directory: %s", clone_dir)
            except Exception as e:
                log.debug("Error removing repository installation directory %s: %s", clone_dir, util.unicodify(e))
    repository.error_message = error_message
    repository.status = status
    repository.deleted = deleted
    repository.uninstalled = uninstalled

    session = app.install_model.context
    session.add(repository)
    with transaction(session):
        session.commit()


__all__ = (
    "check_for_updates",
    "create_or_update_tool_shed_repository",
    "extract_components_from_tuple",
    "generate_tool_shed_repository_install_dir",
    "get_absolute_path_to_file_in_repository",
    "get_ids_of_tool_shed_repositories_being_installed",
    "get_installed_repository",
    "get_installed_tool_shed_repository",
    "get_prior_import_or_install_required_dict",
    "get_repo_info_tuple_contents",
    "get_repository_admin_role_name",
    "get_repository_and_repository_dependencies_from_repo_info_dict",
    "get_repository_by_id",
    "get_repository_by_name",
    "get_repository_by_name_and_owner",
    "get_repository_dependency_types",
    "get_repository_for_dependency_relationship",
    "get_repository_ids_requiring_prior_import_or_install",
    "get_repository_owner",
    "get_repository_owner_from_clone_url",
    "get_repository_query",
    "get_role_by_id",
    "get_tool_shed_from_clone_url",
    "get_tool_shed_repository_by_id",
    "get_tool_shed_status_for_installed_repository",
    "is_tool_shed_client",
    "repository_was_previously_installed",
    "set_repository_attributes",
)



import threading
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Optional,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

from typing_extensions import Protocol

from galaxy.model.base import ModelMapping
from galaxy.model.tool_shed_install import HasToolBox
from galaxy.security.idencoding import IdEncodingHelper
from galaxy.tool_shed.cache import ToolShedRepositoryCache
from galaxy.tool_util.data import (
    OutputDataset,
    ToolDataTableManager,
)
from galaxy.tool_util.toolbox.base import AbstractToolBox

if TYPE_CHECKING:
    import galaxy.tool_shed.metadata.installed_repository_manger


class DataManagerInterface(Protocol):
    GUID_TYPE: str = "data_manager"
    DEFAULT_VERSION: str = "0.0.1"

    def process_result(self, out_data): ...

    def write_bundle(self, out: Dict[str, OutputDataset]) -> Dict[str, OutputDataset]: ...


class DataManagersInterface(Protocol):
    @property
    def _reload_count(self) -> int: ...

    def load_manager_from_elem(
        self, data_manager_elem, tool_path=None, add_manager=True
    ) -> Optional[DataManagerInterface]: ...

    def get_manager(self, data_manager_id: str) -> Optional[DataManagerInterface]: ...

    def remove_manager(self, manager_ids: Union[str, List[str]]) -> None: ...


ToolBoxType = TypeVar("ToolBoxType", bound="AbstractToolBox")


class InstallationTarget(HasToolBox, Generic[ToolBoxType]):
    data_managers: DataManagersInterface
    install_model: ModelMapping
    model: ModelMapping
    security: IdEncodingHelper
    config: Any
    installed_repository_manager: "galaxy.tool_shed.metadata.installed_repository_manger.InstalledRepositoryManager"
    watchers: Any  # TODO: interface...
    _toolbox_lock: threading.RLock
    tool_shed_repository_cache: Optional[ToolShedRepositoryCache]
    tool_data_tables: ToolDataTableManager

    def wait_for_toolbox_reload(self, old_toolbox: ToolBoxType) -> None: ...



"""
Manager and Serializers for Visualizations.

Visualizations are saved configurations/variables used to
reproduce a specific view in a Galaxy visualization.
"""

import logging
from typing import (
    Dict,
    List,
    Tuple,
)

from sqlalchemy import (
    false,
    or_,
    true,
)
from sqlalchemy.orm import aliased

from galaxy import (
    exceptions,
    model,
)
from galaxy.managers import (
    base,
    sharable,
)
from galaxy.managers.context import ProvidesUserContext
from galaxy.model.index_filter_util import (
    append_user_filter,
    raw_text_column_filter,
    tag_filter,
    text_column_filter,
)
from galaxy.schema.visualization import VisualizationIndexQueryPayload
from galaxy.structured_app import MinimalManagerApp
from galaxy.util.search import (
    FilteredTerm,
    parse_filters_structured,
    RawTextTerm,
)

log = logging.getLogger(__name__)


INDEX_SEARCH_FILTERS = {
    "title": "title",
    "slug": "slug",
    "tag": "tag",
    "user": "user",
    "u": "user",
    "s": "slug",
    "t": "tag",
    "is": "is",
}


class VisualizationManager(sharable.SharableModelManager):
    """
    Handle operations outside and between visualizations and other models.
    """

    # TODO: copy, revisions

    model_class = model.Visualization
    foreign_key_name = "visualization"
    user_share_model = model.VisualizationUserShareAssociation

    tag_assoc = model.VisualizationTagAssociation
    annotation_assoc = model.VisualizationAnnotationAssociation
    rating_assoc = model.VisualizationRatingAssociation

    def index_query(
        self, trans: ProvidesUserContext, payload: VisualizationIndexQueryPayload, include_total_count: bool = False
    ) -> Tuple[List[model.Visualization], int]:
        show_deleted = payload.deleted
        show_own = payload.show_own
        show_published = payload.show_published
        show_shared = payload.show_shared
        is_admin = trans.user_is_admin
        user = trans.user

        if not user and not show_published:
            message = "Requires user to log in."
            raise exceptions.RequestParameterInvalidException(message)

        query = trans.sa_session.query(self.model_class)

        filters = []
        if show_own or (not show_published and not show_shared and not is_admin):
            filters = [self.model_class.user == user]
        if show_published:
            filters.append(self.model_class.published == true())
        if user and show_shared:
            filters.append(self.user_share_model.user == user)
            query = query.outerjoin(self.model_class.users_shared_with)
        query = query.filter(or_(*filters))

        if payload.user_id:
            query = query.filter(self.model_class.user_id == payload.user_id)

        if payload.search:
            search_query = payload.search
            parsed_search = parse_filters_structured(search_query, INDEX_SEARCH_FILTERS)

            def p_tag_filter(term_text: str, quoted: bool):
                nonlocal query
                alias = aliased(model.VisualizationTagAssociation)
                query = query.outerjoin(self.model_class.tags.of_type(alias))
                return tag_filter(alias, term_text, quoted)

            for term in parsed_search.terms:
                if isinstance(term, FilteredTerm):
                    key = term.filter
                    q = term.text
                    if key == "tag":
                        pg = p_tag_filter(term.text, term.quoted)
                        query = query.filter(pg)
                    elif key == "title":
                        query = query.filter(text_column_filter(self.model_class.title, term))
                    elif key == "slug":
                        query = query.filter(text_column_filter(self.model_class.slug, term))
                    elif key == "user":
                        query = append_user_filter(query, self.model_class, term)
                    elif key == "is":
                        if q == "deleted":
                            show_deleted = True
                        if q == "published":
                            query = query.filter(self.model_class.published == true())
                        if q == "importable":
                            query = query.filter(self.model_class.importable == true())
                        elif q == "shared_with_me":
                            if not show_shared:
                                message = "Can only use tag is:shared_with_me if show_shared parameter also true."
                                raise exceptions.RequestParameterInvalidException(message)
                            query = query.filter(self.user_share_model.user == user)
                elif isinstance(term, RawTextTerm):
                    tf = p_tag_filter(term.text, False)
                    alias = aliased(model.User)
                    query = query.outerjoin(self.model_class.user.of_type(alias))
                    query = query.filter(
                        raw_text_column_filter(
                            [
                                self.model_class.title,
                                self.model_class.slug,
                                tf,
                                alias.username,
                            ],
                            term,
                        )
                    )

        if (show_published or show_shared) and not is_admin:
            show_deleted = False

        query = query.filter(self.model_class.deleted == (true() if show_deleted else false())).distinct()

        if include_total_count:
            total_matches = query.count()
        else:
            total_matches = None
        sort_column = getattr(model.Visualization, payload.sort_by)
        if payload.sort_desc:
            sort_column = sort_column.desc()
        query = query.order_by(sort_column)
        if payload.limit is not None:
            query = query.limit(payload.limit)
        if payload.offset is not None:
            query = query.offset(payload.offset)
        return query, total_matches


class VisualizationSerializer(sharable.SharableModelSerializer):
    """
    Interface/service object for serializing visualizations into dictionaries.
    """

    model_manager_class = VisualizationManager
    SINGLE_CHAR_ABBR = "v"

    def __init__(self, app: MinimalManagerApp):
        super().__init__(app)
        self.visualization_manager = self.manager

        self.default_view = "summary"
        self.add_view("summary", [])
        self.add_view("detailed", [])

    def add_serializers(self):
        super().add_serializers()
        serializers: Dict[str, base.Serializer] = {}
        self.serializers.update(serializers)


class VisualizationDeserializer(sharable.SharableModelDeserializer):
    """
    Interface/service object for validating and deserializing
    dictionaries into visualizations.
    """

    model_manager_class = VisualizationManager

    def __init__(self, app):
        super().__init__(app)
        self.visualization_manager = self.manager

    def add_deserializers(self):
        super().add_deserializers()
        self.deserializers.update({})
        self.deserializable_keyset.update(self.deserializers.keys())



from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
)
from typing_extensions import (
    Literal,
    TypedDict,
)


class Repository(BaseModel):
    # element/collection view on the backend have same keys/impl
    id: str
    name: str
    owner: str
    type: str  # TODO: enum
    remote_repository_url: Optional[str] = None
    homepage_url: Optional[str] = None
    description: str
    user_id: str
    private: bool
    deleted: bool
    times_downloaded: int
    deprecated: bool
    create_time: str
    update_time: str


class DetailedRepository(Repository):
    long_description: Optional[str]


class RepositoryPermissions(BaseModel):
    allow_push: List[str]
    can_manage: bool  # can the requesting user manage the repository
    can_push: bool


class RepositoryRevisionReadmes(RootModel):
    root: Dict[str, str]


class CreateUserRequest(BaseModel):
    username: str
    email: str
    password: str


class User(BaseModel):
    id: str
    username: str


class UserV2(User):
    is_admin: bool


class Category(BaseModel):
    id: str
    name: str
    description: str
    deleted: bool
    repositories: int


class CreateCategoryRequest(BaseModel):
    name: str
    description: Optional[str] = None


class ValidRepostiroyUpdateMessage(BaseModel):
    message: str


class FailedRepositoryUpdateMessage(BaseModel):
    err_msg: str


class GetOrderedInstallableRevisionsRequest(BaseModel):
    name: str
    owner: str


class OrderedInstallableRevisions(RootModel):
    root: List[str]


RepositoryType = Literal[
    "repository_suite_definition",
    "tool_dependency_definition",
    "unrestricted",
]


class CreateRepositoryRequest(BaseModel):
    name: str
    synopsis: str
    description: Optional[str] = None
    remote_repository_url: Optional[str] = None
    homepage_url: Optional[str] = None
    type_: RepositoryType = Field(
        "unrestricted",
        alias="type",
        title="Type",
    )
    category_ids: Optional[Union[List[str], str]] = Field(
        ...,
        alias="category_ids[]",
        title="Category IDs",
    )
    model_config = ConfigDict(populate_by_name=True)


class RepositoryUpdateRequest(BaseModel):
    commit_message: Optional[str] = None


class RepositoryUpdate(RootModel):
    root: Union[ValidRepostiroyUpdateMessage, FailedRepositoryUpdateMessage]

    @property
    def is_ok(self):
        return isinstance(self.root, ValidRepostiroyUpdateMessage)


class RepositoryTool(BaseModel):
    # Added back in post v2 in order for the frontend to render
    # tool descriptions on the repository page.
    description: str
    guid: str
    id: str
    name: str
    requirements: list
    tool_config: str
    tool_type: str
    version: str
    # add_to_tool_panel: bool
    # tests: list
    # version_string_cmd: Optional[str]


class RepositoryRevisionMetadata(BaseModel):
    id: str
    repository: Repository
    repository_dependencies: List["RepositoryDependency"]
    tools: Optional[List["RepositoryTool"]] = None
    invalid_tools: List[str]  # added for rendering list of invalid tools in 2.0 frontend
    repository_id: str
    numeric_revision: int
    changeset_revision: str
    malicious: bool
    downloadable: bool
    missing_test_components: bool
    has_repository_dependencies: bool
    includes_tools: bool
    includes_tools_for_display_in_tool_panel: bool
    # Deprecate these...
    includes_tool_dependencies: Optional[bool] = None
    includes_datatypes: Optional[bool] = None
    includes_workflows: Optional[bool] = None


class RepositoryDependency(RepositoryRevisionMetadata):
    # This only needs properties for tests it seems?
    # e.g. test_0550_metadata_updated_dependencies.py
    pass


class RepositoryMetadata(RootModel):
    root: Dict[str, RepositoryRevisionMetadata]

    @property
    def latest_revision(self) -> RepositoryRevisionMetadata:
        return list(self.root.values())[-1]

    @property
    def tip(self) -> str:
        if self.is_new:
            return "000000000000"
        else:
            return self.latest_revision.changeset_revision

    @property
    def is_new(self) -> bool:
        return len(self.root) == 0


class ResetMetadataOnRepositoryRequest(BaseModel):
    repository_id: str


class ResetMetadataOnRepositoryResponse(BaseModel):
    status: str  # TODO: enum...
    repository_status: List[str]
    start_time: str
    stop_time: str


class ToolSearchRequest(BaseModel):
    q: str
    page: Optional[int] = None
    page_size: Optional[int] = None


class ToolSearchHitTool(BaseModel):
    id: str
    repo_owner_username: str
    repo_name: str
    name: str
    description: str


class ToolSearchHit(BaseModel):
    tool: ToolSearchHitTool
    matched_terms: Dict[str, Any]
    score: float


class ToolSearchResults(BaseModel):
    # These next three really should be ints :<
    total_results: str
    page: str
    page_size: str
    hostname: str
    hits: List[ToolSearchHit]

    def find_search_hit(self, repository: Repository) -> Optional[ToolSearchHit]:
        matching_hit: Optional[ToolSearchHit] = None

        for hit in self.hits:
            owner_matches = hit.tool.repo_owner_username == repository.owner
            name_matches = hit.tool.repo_name == repository.name
            if owner_matches and name_matches:
                matching_hit = hit
                break

        return matching_hit


class RepositoryIndexRequest(BaseModel):
    owner: Optional[str] = None
    name: Optional[str] = None
    deleted: str = "false"


class RepositoriesByCategory(BaseModel):
    id: str
    name: str
    description: str
    repository_count: int
    repositories: List[Repository]


class RepositoryIndexResponse(RootModel):
    root: List[Repository]


class RepositorySearchRequest(BaseModel):
    q: str
    page: Optional[int] = None
    page_size: Optional[int] = None


class RepositorySearchResult(BaseModel):
    id: str
    name: str
    repo_owner_username: str
    description: str
    long_description: Optional[str] = None
    remote_repository_url: Optional[str] = None
    homepage_url: Optional[str] = None
    last_update: Optional[str] = None
    full_last_updated: str
    repo_lineage: str
    approved: bool
    times_downloaded: int
    categories: str


class RepositorySearchHit(BaseModel):
    score: float
    repository: RepositorySearchResult


class RepositorySearchResults(BaseModel):
    total_results: str
    page: str
    page_size: str
    hostname: str
    hits: List[RepositorySearchHit]


class GetInstallInfoRequest(BaseModel):
    owner: str
    name: str
    changeset_revision: str


class ValidToolDict(TypedDict):
    add_to_tool_panel: bool
    description: str
    guid: str
    id: str
    name: str
    requirements: list
    tests: list
    tool_config: str
    tool_type: str
    version: str
    version_string_cmd: Optional[str]


class RepositoryMetadataInstallInfoDict(TypedDict):
    changeset_revision: str
    downloadable: bool
    has_repository_dependencies: bool
    has_repository_dependencies_only_if_compiling_contained_td: bool
    id: str
    includes_datatypes: bool
    includes_tool_dependencies: bool
    includes_tools: bool
    includes_tools_for_display_in_tool_panel: bool
    includes_workflows: bool
    malicious: bool
    repository_id: str
    url: str
    valid_tools: List[ValidToolDict]


# So hard to type this... the keys are repo names and the elements
# are tuples that have been list-ified.
ExtraRepoInfo = Dict[str, List]
# {
#     "add_column": [
#         "add_column hello",
#         "http://test@localhost:9009/repos/test/add_column",
#         "3a08cc21466f",
#         "1",
#         "test",
#         {},
#         {}
#     ]
# }


class EmptyDict(TypedDict):
    pass


LegacyInstallInfoTuple = Tuple[
    Optional[Dict], Union[RepositoryMetadataInstallInfoDict, EmptyDict], Union[ExtraRepoInfo, EmptyDict]
]


class RepositoryExtraInstallInfo(BaseModel):
    name: str
    description: str
    repository_clone_url: str
    changeset_revision: str
    ctx_rev: str
    repository_owner: str
    repository_dependencies: Optional[Dict] = None
    # tool dependencies not longer work so don't transmit them in v2?
    # tool_dependencies: Optional[Dict]

    @staticmethod
    def from_legacy_dict(as_dict: ExtraRepoInfo) -> "RepositoryExtraInstallInfo":
        assert len(as_dict) == 1
        repo_name = next(iter(as_dict.keys()))
        info_indexable = as_dict[repo_name]
        return RepositoryExtraInstallInfo(
            name=repo_name,
            description=info_indexable[0],
            repository_clone_url=info_indexable[1],
            changeset_revision=info_indexable[2],
            ctx_rev=info_indexable[3],
            repository_owner=info_indexable[4],
            repository_dependencies=info_indexable[5],
        )


class ValidTool(BaseModel):
    add_to_tool_panel: bool
    description: str
    guid: str
    id: str
    name: str
    requirements: list
    tests: list
    tool_config: str
    tool_type: str
    version: str
    version_string_cmd: Optional[str] = None

    @staticmethod
    def from_legacy_dict(as_dict: ValidToolDict) -> "ValidTool":
        return ValidTool(**as_dict)

    @staticmethod
    def from_legacy_list(as_dicts: List[ValidToolDict]) -> List["ValidTool"]:
        return list(ValidTool.from_legacy_dict(d) for d in as_dicts)


class RepositoryMetadataInstallInfo(BaseModel):
    id: str
    changeset_revision: str
    downloadable: bool
    has_repository_dependencies: bool
    includes_tools: bool
    includes_tools_for_display_in_tool_panel: bool
    malicious: bool
    repository_id: str
    url: str
    valid_tools: List[ValidTool]
    # no longer used, don't transmit.
    # has_repository_dependencies_only_if_compiling_contained_td: bool
    # includes_datatypes: bool
    # includes_tool_dependencies: bool
    # includes_workflows: bool

    @staticmethod
    def from_legacy_dict(as_dict: RepositoryMetadataInstallInfoDict) -> "RepositoryMetadataInstallInfo":
        return RepositoryMetadataInstallInfo(
            id=as_dict["id"],
            changeset_revision=as_dict["changeset_revision"],
            downloadable=as_dict["downloadable"],
            has_repository_dependencies=as_dict["has_repository_dependencies"],
            includes_tools=as_dict["includes_tools"],
            includes_tools_for_display_in_tool_panel=as_dict["includes_tools_for_display_in_tool_panel"],
            malicious=as_dict["malicious"],
            repository_id=as_dict["repository_id"],
            url=as_dict["url"],
            valid_tools=ValidTool.from_legacy_list(as_dict.get("valid_tools", [])),
        )


class InstallInfo(BaseModel):
    metadata_info: Optional[RepositoryMetadataInstallInfo] = None
    repo_info: Optional[RepositoryExtraInstallInfo] = None


def from_legacy_install_info(legacy_install_info: LegacyInstallInfoTuple) -> InstallInfo:
    repo_metadata_install_info: Union[RepositoryMetadataInstallInfoDict, EmptyDict]
    extra_info: Union[ExtraRepoInfo, EmptyDict]
    _, repo_metadata_install_info, extra_info = legacy_install_info
    if repo_metadata_install_info:
        metadata_info = RepositoryMetadataInstallInfo.from_legacy_dict(
            cast(RepositoryMetadataInstallInfoDict, repo_metadata_install_info)
        )
    else:
        metadata_info = None
    if extra_info:
        repo_info = RepositoryExtraInstallInfo.from_legacy_dict(extra_info)
    else:
        repo_info = None
    return InstallInfo(
        metadata_info=metadata_info,
        repo_info=repo_info,
    )


class BuildSearchIndexResponse(BaseModel):
    repositories_indexed: int
    tools_indexed: int


class Version(BaseModel):
    version_major: str
    version: str
    api_version: str = "v1"



def exec_after_process(app, inp_data, out_data, param_dict, tool, stdout, stderr):
    output_data = next(iter(out_data.values()))
    new_stdout = ""
    split_stdout = stdout.split("\n")
    for line in split_stdout:
        if line.startswith("#FILE1"):
            fields = line.split("\t")
            dbkey = fields[1]
            output_data.dbkey = dbkey
            output_data.name = "%s (%s)" % (output_data.name, dbkey)
            app.model.context.add(output_data)
            app.model.context.flush()
        else:
            new_stdout = "%s\n%s" % (new_stdout, line)
    for data in output_data.creating_job.output_datasets:
        data = data.dataset
        data.info = new_stdout
        app.model.context.add(data)
        app.model.context.flush()



import logging
import os

from galaxy import util
from galaxy.tool_shed.galaxy_install.installed_repository_manager import InstalledRepositoryManager
from galaxy.tool_shed.util import utility_container_manager
from galaxy.util import UNKNOWN
from galaxy.util.tool_shed.common_util import parse_repository_dependency_tuple

log = logging.getLogger(__name__)


class DependencyDisplayer:
    def __init__(self, app):
        self.app = app

    def _add_installation_directories_to_tool_dependencies(self, tool_dependencies):
        """
        Determine the path to the installation directory for each of the received
        tool dependencies.  This path will be displayed within the tool dependencies
        container on the select_tool_panel_section or reselect_tool_panel_section
        pages when installing or reinstalling repositories that contain tools with
        the defined tool dependencies.  The list of tool dependencies may be associated
        with more than a single repository.
        """
        for dependency_key, requirements_dict in tool_dependencies.items():
            if dependency_key in ["set_environment"]:
                continue
            repository_name = requirements_dict.get("repository_name", UNKNOWN)
            repository_owner = requirements_dict.get("repository_owner", UNKNOWN)
            changeset_revision = requirements_dict.get("changeset_revision", UNKNOWN)
            dependency_name = requirements_dict["name"]
            version = requirements_dict["version"]
            if self.app.tool_dependency_dir:
                root_dir = self.app.tool_dependency_dir
            else:
                root_dir = "<set your tool_dependency_dir in your Galaxy configuration file>"
            install_dir = os.path.join(
                root_dir, dependency_name, version, repository_owner, repository_name, changeset_revision
            )
            requirements_dict["install_dir"] = install_dir
            tool_dependencies[dependency_key] = requirements_dict
        return tool_dependencies

    def generate_message_for_invalid_repository_dependencies(self, metadata_dict, error_from_tuple=False):
        """
        Get or generate and return an error message associated with an invalid repository dependency.
        """
        message = ""
        if metadata_dict:
            if error_from_tuple:
                # Return the error messages associated with a set of one or more invalid repository
                # dependency tuples.
                invalid_repository_dependencies_dict = metadata_dict.get("invalid_repository_dependencies", None)
                if invalid_repository_dependencies_dict is not None:
                    invalid_repository_dependencies = invalid_repository_dependencies_dict.get(
                        "invalid_repository_dependencies", []
                    )
                    for repository_dependency_tup in invalid_repository_dependencies:
                        (
                            toolshed,
                            name,
                            owner,
                            changeset_revision,
                            prior_installation_required,
                            only_if_compiling_contained_td,
                            error,
                        ) = parse_repository_dependency_tuple(repository_dependency_tup, contains_error=True)
                        if error:
                            message += f"{error}  "
            else:
                # The complete dependency hierarchy could not be determined for a repository being installed into
                # Galaxy.  This is likely due to invalid repository dependency definitions, so we'll get them from
                # the metadata and parse them for display in an error message.  This will hopefully communicate the
                # problem to the user in such a way that a resolution can be determined.
                message += (
                    "The complete dependency hierarchy could not be determined for this repository, so no required "
                )
                message += "repositories will not be installed.  This is likely due to invalid repository dependency definitions.  "
                repository_dependencies_dict = metadata_dict.get("repository_dependencies", None)
                if repository_dependencies_dict is not None:
                    rd_tups = repository_dependencies_dict.get("repository_dependencies", None)
                    if rd_tups is not None:
                        message += "Here are the attributes of the dependencies defined for this repository to help determine the "
                        message += "cause of this problem.<br/>"
                        message += '<table cellpadding="2" cellspacing="2">'
                        message += (
                            "<tr><th>Tool shed</th><th>Repository name</th><th>Owner</th><th>Changeset revision</th>"
                        )
                        message += "<th>Prior install required</th></tr>"
                        for rd_tup in rd_tups:
                            (
                                tool_shed,
                                name,
                                owner,
                                changeset_revision,
                                pir,
                                oicct,
                            ) = parse_repository_dependency_tuple(rd_tup)
                            if util.asbool(pir):
                                pir_str = "True"
                            else:
                                pir_str = ""
                            message += f"<tr><td>{tool_shed}</td><td>{name}</td><td>{owner}</td><td>{changeset_revision}</td><td>{pir_str}</td></tr>"
                        message += "</table>"
        return message

    def generate_message_for_invalid_tool_dependencies(self, metadata_dict):
        """
        Tool dependency definitions can only be invalid if they include a definition for a complex
        repository dependency and the repository dependency definition is invalid.  This method
        retrieves the error message associated with the invalid tool dependency for display in the
        caller.
        """
        message = ""
        if metadata_dict:
            invalid_tool_dependencies = metadata_dict.get("invalid_tool_dependencies", None)
            if invalid_tool_dependencies:
                for requirement_dict in invalid_tool_dependencies.values():
                    error = requirement_dict.get("error", None)
                    if error:
                        message = f"{error}  "
        return message

    def generate_message_for_orphan_tool_dependencies(self, repository, metadata_dict):
        """
        The designation of a ToolDependency into the "orphan" category has evolved over time,
        and is significantly restricted since the introduction of the TOOL_DEPENDENCY_DEFINITION
        repository type.  This designation is still critical, however, in that it handles the
        case where a repository contains both tools and a tool_dependencies.xml file, but the
        definition in the tool_dependencies.xml file is in no way related to anything defined
        by any of the contained tool's requirements tag sets.  This is important in that it is
        often a result of a typo (e.g., dependency name or version) that differs between the tool
        dependency definition within the tool_dependencies.xml file and what is defined in the
        tool config's <requirements> tag sets.  In these cases, the user should be presented with
        a warning message, and this warning message is is in fact displayed if the following
        is_orphan attribute is True.  This is tricky because in some cases it may be intentional,
        and tool dependencies that are categorized as "orphan" are in fact valid.
        """
        has_orphan_package_dependencies = False
        has_orphan_set_environment_dependencies = False
        message = ""
        package_orphans_str = ""
        set_environment_orphans_str = ""
        # Tool dependencies are categorized as orphan only if the repository contains tools.
        if metadata_dict:
            tools = metadata_dict.get("tools", [])
            tool_dependencies = metadata_dict.get("tool_dependencies", {})
            # The use of the orphan_tool_dependencies category in metadata has been deprecated,
            # but we still need to check in case the metadata is out of date.
            orphan_tool_dependencies = metadata_dict.get("orphan_tool_dependencies", {})
            # Updating should cause no problems here since a tool dependency cannot be included
            # in both dictionaries.
            tool_dependencies.update(orphan_tool_dependencies)
            if tool_dependencies and tools:
                for td_key, requirements_dict in tool_dependencies.items():
                    if td_key == "set_environment":
                        # "set_environment": [{"name": "R_SCRIPT_PATH", "type": "set_environment"}]
                        for env_requirements_dict in requirements_dict:
                            name = env_requirements_dict["name"]
                            type = env_requirements_dict["type"]
                            if self._tool_dependency_is_orphan(type, name, None, tools):
                                if not has_orphan_set_environment_dependencies:
                                    has_orphan_set_environment_dependencies = True
                                set_environment_orphans_str += f"<b>* name:</b> {name}, <b>type:</b> {type}<br/>"
                    else:
                        # "R/2.15.1": {"name": "R", "readme": "some string", "type": "package", "version": "2.15.1"}
                        name = requirements_dict["name"]
                        type = requirements_dict["type"]
                        version = requirements_dict["version"]
                        if self._tool_dependency_is_orphan(type, name, version, tools):
                            if not has_orphan_package_dependencies:
                                has_orphan_package_dependencies = True
                            package_orphans_str += (
                                f"<b>* name:</b> {name}, <b>type:</b> {type}, <b>version:</b> {version}<br/>"
                            )
        if has_orphan_package_dependencies:
            message += "The settings for <b>name</b>, <b>version</b> and <b>type</b> from a "
            message += "contained tool configuration file's <b>requirement</b> tag does not match "
            message += "the information for the following tool dependency definitions in the "
            message += "<b>tool_dependencies.xml</b> file, so these tool dependencies have no "
            message += "relationship with any tools within this repository.<br/>"
            message += package_orphans_str
        if has_orphan_set_environment_dependencies:
            message += "The settings for <b>name</b> and <b>type</b> from a contained tool "
            message += "configuration file's <b>requirement</b> tag does not match the information "
            message += "for the following tool dependency definitions in the <b>tool_dependencies.xml</b> "
            message += "file, so these tool dependencies have no relationship with any tools within "
            message += "this repository.<br/>"
            message += set_environment_orphans_str
        return message

    def populate_containers_dict_from_repository_metadata(self, repository):
        """
        Retrieve necessary information from the received repository's metadata to populate the
        containers_dict for display.  This method is called only from Galaxy (not the tool shed)
        when displaying repository dependencies for installed repositories and when displaying
        them for uninstalled repositories that are being reinstalled.
        """
        if metadata := repository.metadata_:
            irm = InstalledRepositoryManager(self.app)
            # Handle repository dependencies.
            (
                installed_repository_dependencies,
                missing_repository_dependencies,
            ) = irm.get_installed_and_missing_repository_dependencies(repository)
            # Handle the current repository's tool dependencies.
            repository_tool_dependencies = metadata.get("tool_dependencies", None)
            # Make sure to display missing tool dependencies as well.
            repository_invalid_tool_dependencies = metadata.get("invalid_tool_dependencies", None)
            if repository_invalid_tool_dependencies is not None:
                if repository_tool_dependencies is None:
                    repository_tool_dependencies = {}
                repository_tool_dependencies.update(repository_invalid_tool_dependencies)
            gucm = GalaxyUtilityContainerManager(self.app)
            containers_dict = gucm.build_repository_containers(
                missing_repository_dependencies=missing_repository_dependencies,
                repository_dependencies=installed_repository_dependencies,
            )
        else:
            containers_dict = dict(
                repository_dependencies=None,
            )
        return containers_dict

    def _tool_dependency_is_orphan(self, type, name, version, tools):
        """
        Determine if the combination of the received type, name and version is defined in the <requirement>
        tag for at least one tool in the received list of tools.  If not, the tool dependency defined by the
        combination is considered an orphan in its repository in the tool shed.
        """
        if type == "package":
            if name and version:
                for tool_dict in tools:
                    requirements = tool_dict.get("requirements", [])
                    for requirement_dict in requirements:
                        req_name = requirement_dict.get("name", None)
                        req_version = requirement_dict.get("version", None)
                        req_type = requirement_dict.get("type", None)
                        if req_name == name and req_version == version and req_type == type:
                            return False
        elif type == "set_environment":
            if name:
                for tool_dict in tools:
                    requirements = tool_dict.get("requirements", [])
                    for requirement_dict in requirements:
                        req_name = requirement_dict.get("name", None)
                        req_type = requirement_dict.get("type", None)
                        if req_name == name and req_type == type:
                            return False
        return True


class GalaxyUtilityContainerManager(utility_container_manager.UtilityContainerManager):
    def __init__(self, app):
        self.app = app

    def build_repository_containers(
        self,
        missing_repository_dependencies,
        repository_dependencies,
    ):
        """
        Return a dictionary of containers for the received repository's dependencies and readme files for
        display during installation to Galaxy.
        """
        containers_dict = dict(
            repository_dependencies=None,
            missing_repository_dependencies=None,
        )
        # Some of the tool dependency folders will include links to display tool dependency information, and
        # some of these links require the repository id.  However we need to be careful because sometimes the
        # repository object is None.
        try:
            folder_id = 0
            # Installed repository dependencies container.
            if repository_dependencies:
                label = "Installed repository dependencies"
                folder_id, repository_dependencies_root_folder = self.build_repository_dependencies_folder(
                    folder_id=folder_id, repository_dependencies=repository_dependencies, label=label, installed=True
                )
                containers_dict["repository_dependencies"] = repository_dependencies_root_folder
            # Missing repository dependencies container.
            if missing_repository_dependencies:
                folder_id, missing_repository_dependencies_root_folder = self.build_repository_dependencies_folder(
                    folder_id=folder_id,
                    repository_dependencies=missing_repository_dependencies,
                    label="Missing repository dependencies",
                    installed=False,
                )
                containers_dict["missing_repository_dependencies"] = missing_repository_dependencies_root_folder
        except Exception as e:
            log.debug(f"Exception in build_repository_containers: {str(e)}")
        return containers_dict


def build_manage_repository_dict(app, status, repository):
    dd = DependencyDisplayer(app)
    containers_dict = dd.populate_containers_dict_from_repository_metadata(
        repository=repository,
    )
    management_dict = {
        "status": status,
    }
    missing_repo_dependencies = containers_dict.get("missing_repository_dependencies", None)
    if missing_repo_dependencies:
        management_dict["missing_repository_dependencies"] = missing_repo_dependencies.to_dict()
    repository_dependencies = containers_dict.get("repository_dependencies", None)
    if repository_dependencies:
        management_dict["repository_dependencies"] = repository_dependencies.to_dict()
    return management_dict
